/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdarg.h>
#include "gfx.h"

#include "font.h"
#include "ptyutil.h"
#include "util.h"
#include "matrix.h"
#include "msg.h"
#include "mconst.h"
#include "ifad.h"

static int   wd_cols= 80;
static int   wd_rows= 24;
static int   exit_event_loop;
static int   click_x=-1, click_y=-1, click_at;
static int   imode;

static int   cmd_arg_fsize= 16;
static int   cmd_arg_wextra= 1;
static float cmd_arg_hscale= 1.0;

static int active_tab=-1;

enum { FD_SERVER= 100, FD_DATA };
static struct mchan  mc_in, mc_out;
static int           ro_server, ro_sock;

static int   CURSORx=100, CURSORy=100;
static int   rem_w, rem_h;     // remote width and height
static int   tracking_cursor;

static void deactivate(int tn);
static void activate(int tn);
static int start_tab(int tn);
static void update_cursor(int Ux0, int Uy0, int Uw, int Uh);
static void move_cursor(int x, int y);

void selection_lost();

static void msg_out(int type, int ni, int ns, void (*encf)(struct msg*), ...)
{
  struct msg M;
  va_list ap;
  int i;
  M.type= type;
  M.ni= ni;
  M.ns= ns;
  M.encoded= encf;
  va_start(ap, encf);
  for(i=0;i<ni;i++)
    M.iarg[i]= va_arg(ap, int);
  for(i=0;i<ns;i++)
  {
    M.larg[i]= va_arg(ap, int);
    M.sarg[i]= va_arg(ap, char*);
    if (M.larg[i]<0) M.larg[i]= strlen(M.sarg[i])+1;
  }
  va_end(ap);
  gfx_stop_monitor_file(-1, FD_DATA);
  gfx_monitor_file(ro_sock, FD_DATA, 1, 1, 1);
  mchan_queue(&mc_out, &M);
}

typedef struct 
{
  int bg, fg, og;
  unsigned int normal[256], bold[256];
} color_t;

static color_t colors[9]=
 {
   { 0x000000, 0xaaaaaa, 0xffffff, },
   { 0x00004b, 0xaaaaaa, 0xffffff, },
   { 0x83713e, 0x000000, 0xffffff, },
   { 0x380000, 0xffffff, 0xffff00, },
   { 0x103164, 0xffffff, 0x00ffff, },
   { 0x004000, 0xffffff, 0xffff00, },
   { 0x404040, 0xffffff, 0xff0000, },
   { 0xcccccc, 0x000000, 0x8c4422, },
   { 0x00ffff, 0x000000, 0x000000, }
 };
static int curcolor;


static screen_t *SCR;

static char **child_argv;

typedef struct
{
  int present;
  char *str;
  screen_t *screen;
} tab_t;

static tab_t tabs[9];

static void mouse_move(int x,int y);
static void mouse_release(int button, int x, int y);
static void mouse_press(int button, int x, int y);

static void write_tty_str(char *b)
{
  if (!b || active_tab<0) return ;
  msg_out(ROME_SCR2TTY, 1, 1, NULL, active_tab, -1, b);
}

static void setselok(struct msg *M)
{
  free(M->sarg[0]);
}

void have_selection(uint32_t *wstr)
{
  if (!wstr || !wstr[0])
  {
     selection_lost();
     msg_out(ROME_RESETSEL, 0, 0, NULL);
  }
  else
  {
     char *str;
     int i;
     size_t wlen,slen;
     wlen= wstrlen(wstr);
     str= malloc(wlen*6+1);
     utf8_encode_string(wstr, wlen, (uint8_t*) str, &slen);
     str[slen]= 0;
     msg_out(ROME_SETSEL, 0, 1, setselok, -1, str);
     for(i=0;i<8;i++)
       if (tabs[i].present && tabs[i].screen!=SCR)
          screen_select_none(tabs[i].screen);
  }
  if (wstr) free(wstr);
}

#ifdef MUDEBUG
void debug_trap()
{
  printf("DEBUG TRAP\n");
}
#endif

void key_event(int key, unsigned int wide, char *str)
{
  char *b;
  char bu[2];
  int lapa;
  int i;
  if (imode) { help_key(key,wide,str); return ; }
  lapa= 1;
  b= NULL;
  if (key & GFX_K_CONTROL)
  {
    int z;
    z= key & ~GFX_K_CONTROL;
    if (z>='a' &&  z<='z') 
    {
      bu[0]= z-'a'+1;
      bu[1]= 0;
      b= bu;
      goto done;
    }
    if (z>='1' && z<='8')
    {
      activate(start_tab(z-'1'));
      return ;
    }
#ifdef MUDEBUG
    if (z=='9')
    {
      debug_trap();
      return ;
    }
#endif
  }
  switch(key)
  {
  case GFX_K_CONTROL | GFX_K_RIGHT:
    for(i=1;i<8;i++)
      if (tabs[(active_tab+i)%8].present)
       { activate((active_tab+i)%8); break; }
    break;
  case GFX_K_CONTROL | GFX_K_LEFT:
    for(i=1;i<8;i++)
      if (tabs[(active_tab-i+8)%8].present)
       { activate((active_tab-i+8)%8); break; }
    break;
  
  case GFX_K_SHIFT   | GFX_K_INSERT: msg_out(ROME_REQSEL, 0,0, NULL); break;
  case GFX_K_CONTROL | GFX_K_INSERT: msg_out(ROME_REQCLIP, 0,0, NULL); break;
  case GFX_K_CONTROL | ' ':          msg_out(ROME_SEL2CLIP, 0,0,NULL); break;

  case GFX_K_SHIFT | GFX_K_PAGEUP:   lapa= 0; screen_page_prev(SCR); break;
  case GFX_K_SHIFT | GFX_K_PAGEDOWN: lapa= 0; screen_page_next(SCR); break;

  case GFX_K_PRT:                b= str; break;
  case GFX_K_BACKSPACE:          b= screen_bs_mode(SCR) ? "\177":"\b"; break;
  case GFX_K_DELETE:             b= screen_del_mode(SCR) ?
                                                  "\177" : "\033[3~"; break;
  case GFX_K_RETURN:             b= "\r"; break;
  case GFX_K_TAB:                b="\t"; break;
  case GFX_K_CONTROL | '\\':     b= "\034"; break;
  case GFX_K_UP:                 b= "\033OA"; break;
  case GFX_K_DOWN:               b= "\033OB"; break;
  case GFX_K_LEFT:               b= "\033OD"; break;
  case GFX_K_RIGHT:              b= "\033OC"; break;
  case GFX_K_PAGEUP:             b= "\033[5~"; break;
  case GFX_K_PAGEDOWN:           b= "\033[6~"; break;
  case GFX_K_INSERT:             b= "\033[2~"; break;
  case GFX_K_ESCAPE:             b= "\033"; break;
  case GFX_K_HOME:               b= "\033OH"; break;
  case GFX_K_END:                b= "\033OF"; break;
  default: return ;
  }
done:
  write_tty_str(b);
  if (lapa) screen_page_last(SCR);
}



static void init_colors()
{
  int c;
  for(c=0;c<9;c++)
  {
    gx.blend_user_to_server(colors[c].bg, colors[c].fg, colors[c].normal);
    gx.blend_user_to_server(colors[c].bg, colors[c].og, colors[c].bold);
  }
}

static struct
{
  unsigned int width, height, wpl;
  uint32_t *buffer;
} disp;


static int cro, cco;
static int selb_row=-1, selb_col, sele_row, sele_col;

void set_cursor(int row, int col)
{
  cro= row; cco= col;
}

void set_sel(int br, int bc, int er, int ec)
{
  selb_row= br;
  selb_col= bc;
  sele_row= er;
  sele_col= ec;
}

void display_glyph(int row,int col, uint32_t C)
{
  unsigned int *comap;
  glyph_t *G;
  int ox,oy,x,y;
  unsigned char XO;
  uint32_t chr;

  ox= col*fontW;
  oy= (row+1)*fontH;
  if ((row==cro && col==cco) || (C&CELLATTR_REV)) XO= 255; else XO= 0;

  chr= C&CELL_CHRMASK;
  G= get_glyph(chr);

#define spos(r,c) ((r)*wd_cols+(c))
  if (selb_row!=-1 && spos(selb_row,selb_col)<=spos(row,col) &&
                      spos(sele_row,sele_col)>=spos(row,col))
    comap= colors[8].normal; 
  else if (C&CELLATTR_BOLD)
    comap= colors[curcolor].bold;
  else 
    comap= colors[curcolor].normal;
#undef spos

  for(y=0;y<fontH;y++)
    for(x=0;x<fontW;x++)
      disp.buffer[(y+oy)*disp.wpl+(x+ox)]= comap[G->data[x+y*fontW]^XO];
}

#if 0
static int matrix_dbg=0;

void dump(unsigned char *str,int len)
{
  int i;
  if (matrix_dbg) {
   printf("matrixdbg\n");
    for(i=0;i<len;i++) { if (i%16==0) printf("<\n"); 
       if (str[i]>=' ' && str[i]<127) printf("%c",str[i]); else printf(" %02x ", str[i]); }
     printf("<-----\n");
  }
}
#endif


static int start_tab(int tn)
{
  if (tabs[tn].present) return tn;
  msg_out(ROME_ACTIVATE, 1, 0, NULL, tn);
  tabs[tn].present= 1;
  return tn;
}

static void draw_tab(int tn);

static void deactivate(int tn)
{
  int i;
  // if (!tabs[tn].present) return ;
  /*  012345678 
          X            */
  tabs[tn].present= 0;
  tabs[tn].str[3]= '[';
  tabs[tn].str[5]= ']';
  screen_activate(tabs[tn].screen, 0);

  draw_tab(tn);
  if (tn==active_tab)
  {
    for(i=0;i<8;i++)
     if (tabs[(tn+i)%8].present) break;
    if (i==8) { exit_event_loop=1; return ;}
    activate((tn+i)%8);
  }
}

static void tabmark(int tn, char *mark)
{
  tab_t *T;
  T= tabs + tn;
  T->str[3]= mark[0];
  T->str[5]= mark[1];
  draw_tab(tn);
}

static void activate(int tn)
{
  if (tn==active_tab) return ;
 
  if (active_tab>=0)
  {
    if (tabs[active_tab].present) tabmark(active_tab, "  ");
                             else tabmark(active_tab, "[]");
  }
  SCR= tabs[tn].screen;
  active_tab= tn;
  tabmark(active_tab, "==");
  screen_activate(tabs[tn].screen,1);
  curcolor= tn;
  screen_redraw(SCR);
}


void display_modified()
{
  // gfx_flush();   FIXME: remove
}

void handle_connection()
{
  ro_sock= socket_accept(ro_server);
  if (ro_sock<0)
  {
    perror("accept");
    return ;
  }
  gfx_stop_monitor_file(ro_server, FD_SERVER);
  gfx_monitor_file(ro_sock, FD_DATA, 1, 1, 1);
  mchan_init(&mc_in, ro_sock, 0);
  mchan_init(&mc_out, ro_sock, 1);
  start_tab(0);
}

static void rocheck()
{
  if (ro_sock<0)
    gfx_stop_monitor_file(-1, FD_DATA);
}

static void get_cursor_xy(int *x, int *y, struct msg *M, int o)
{
  if (rem_w) *x= M->iarg[o]*disp.width/rem_w;  else *x= M->iarg[o];
  if (rem_h) *y= M->iarg[o+1]*disp.height/rem_h; else *y= M->iarg[o+1];
}

void handle_data_in()
{
  int x,y;
  struct msg M;
  int tn;
  rocheck();
  if (!mchan_recv(&mc_in, &M)) return ;
  switch(M.type)
  {
  case ROME_KEY: key_event(M.iarg[0], M.iarg[1], M.sarg[0]); break;
  case ROME_TTY2SCR: 
    tn= M.iarg[0];
    if (M.larg[0]>0)
      screen_print_str(tabs[tn].screen,(const uint8_t *) M.sarg[0], M.larg[0]);
    break;
  case ROME_DEACTIVATE: deactivate(M.iarg[0]); break;
  case ROME_PRESENT:
    for(tn=0;tn<8;tn++) 
    {
      tabs[tn].present= (M.iarg[0] >> tn)&1;
      if (tn==active_tab) tabmark(tn,"==");
                     else if (tabs[tn].present) tabmark(tn, "  ");
                                           else tabmark(tn, "[]");
      draw_tab(tn);
    }
    break;
  case ROME_POINTER:
    get_cursor_xy(&x,&y,&M,0);
    move_cursor(x, y);
    if (tracking_cursor)
       mouse_move(CURSORx, CURSORy);
    break;
  case ROME_WINDOWDIM:
    rem_w= M.iarg[0];
    rem_h= M.iarg[1];
    break;
  case ROME_CLICK:
    get_cursor_xy(&x,&y,&M,1);
    mouse_press(M.iarg[0],x, y);
    break;
  case ROME_RELEASE:
    get_cursor_xy(&x,&y,&M,1);
    mouse_release(M.iarg[0],x, y);
    break;
  case ROME_SELCLR:  selection_lost(); break;
  case ROME_SELDATA: write_tty_str(M.sarg[0]); break;
  }
}


void handle_data_out()
{
  rocheck();
  switch(mchan_send(&mc_out))
  {
  case 0: return ;
  case 1: gfx_stop_monitor_file(-1, FD_DATA);
          gfx_monitor_file(ro_sock, FD_DATA, 1, 0, 1);
          break;
  case -1: gfx_stop_monitor_file(-1, FD_DATA);
          break;
  }
}

void handle_data_except()
{
  gfx_stop_monitor_file(-1, FD_DATA);
}

void event_loop()
{
  gfx_event_t ev;
 
  while(!exit_event_loop)
  {
    if (SCR) screen_updates(SCR);
    if (!gfx_event(&ev)) continue;
    switch(ev.type)
    {
    case GFXEV_READY_READ:
      if (ev.key==FD_SERVER) handle_connection();
      else if (ev.key==FD_DATA) handle_data_in();
      break;
    case GFXEV_READY_WRITE:
      handle_data_out();
      break;
    case GFXEV_READY_EXCEPT:
      handle_data_except();
      break;
    }
  }
}

static void mouse_press(int button, int x, int y)
{
  int row;
  if (imode) return ;
  switch(button)
  {
    case 4: // SCROLL UP       FIXME
      if (screen_mode(SCR))
        write_tty_str("\033[5~");
      else
        screen_page_prev(SCR);
      break;
    case 5: // SCROLL DOWN     FIXME
      if (screen_mode(SCR))
        write_tty_str("\033[6~");
      else
	screen_page_next(SCR); 
      break;
    case 1: // LEFT            FIXME
       row= y/fontH-1;
       click_x= x; click_y= y;
       if (row<0)
       {
          click_at= 'B'; 
          return ;
       }
       click_at= 'A';
       tracking_cursor= 1;
       screen_ppress(SCR, click_y/fontH-1, click_x/fontW);
       break;
    case 2:   // MID           FIXME
       msg_out(ROME_REQSEL, 0, 0, NULL);
       break;
  }
}

static void mouse_release(int button, int x, int y)
{
  if (imode) return ;
  switch(button)
  {
    case 1: // LEFT         FIXME
       tracking_cursor= 0;
       if (click_at=='B')
       {
         int T;
         T= click_x/fontW/9;
         if (T<8)
         {
           activate(start_tab(T));
         }
         else
         {
           imode= 'h';
           help_show();
           imode= 0;
           screen_redraw(SCR);
         }
         break;
       }
       screen_prelease(SCR);
       break;
  }
}

static void mouse_move(int x,int y)
{
  if (imode) return ;
  if (click_at=='A')
     screen_pmove(SCR, y/fontH-1, x/fontW);
}

void update_area(int col,int row, int wid, int hei)
{
  int x0, y0, w, h;
  x0= col*fontW, 
  y0= (row+1)*fontH; 
  w= wid*fontW;
  h= hei*fontH;
  update_cursor(x0, y0, w, h);
  gfx_pixgrid_update(Wplane, x0, y0, w, h);
}

void set_title(uint8_t *S)
{
  // gfx_set_window_title(S);
}

void quit_request()
{
  exit_event_loop= 1;
}

void selection_lost()
{
  int i;
  for(i=0;i<8;i++)
    if (tabs[i].present)
      screen_select_none(tabs[i].screen);
}

static void init_tabs()
{
  int i;
  char buf[10];
  for(i=0;i<8;i++)
  {
    memset(buf,' ', 10);
    buf[9]= 0;
    buf[3]= '[';
    buf[4]= i+'1';
    buf[5]= ']';
    tabs[i].str= strdup(buf);
    tabs[i].screen= screen_new(wd_rows, wd_cols);
  }
  tabs[8].str= strdup("  MENU  ");
}

static void draw_tab(int tn)
{
  int j,k;
  int cc;
  uint32_t C;

  cc= curcolor;
  curcolor= tn;
  k= tn*9;
  for(j=0;tabs[tn].str[j];j++,k++)
  {
      C= tabs[tn].str[j];
      display_glyph(-1, k, C);
  }
  update_area(tn*9, -1, 9, 1);
  curcolor= cc;
}

static void draw_bar()
{
  int i,j,k;
  int cc;
  uint32_t C;

  cc= curcolor;
  k= 0;
  C= 0;
  for(i=0;i<9;i++)
  {
    curcolor= i%8;
    for(j=0;tabs[i].str[j];j++,k++)
    {
      C= tabs[i].str[j];
      display_glyph(-1, k, C);
    }
  }
  update_area(0, -1, wd_cols, 1);
  curcolor= cc;
}



static void parse_args(int argc,char **argv)
{
  int i;
  for(i=1;i<argc;i++)
    if (!strcmp(argv[i],"--"))
       break;
  if (i==argc)
  {
    child_argv= argv+1;
    return ;
  }
  child_argv= argv+i+1;
  argc= i;
  for(i=1;i<argc;i++)
  {
    if (!strcmp(argv[i], "-f"))
    {
      if (i==argc-1) 
      {
        fprintf(stderr,"-f needs an argument\n");
        continue;
      }
      cmd_arg_fsize= atoi(argv[i+1]);
      if (cmd_arg_fsize<=0) 
      {
        fprintf(stderr,"font size shall be positive, given %s\n",argv[i+1]);
        cmd_arg_fsize= 10;
      }
      i++;
      continue;
    }
    if (!strcmp(argv[i], "-s"))
    {
      if (i==argc-1) 
      {
        fprintf(stderr,"-s needs an argument\n");
        continue;
      }
      cmd_arg_wextra= atoi(argv[i+1]);
      i++;
      continue;
    }
    if (!strcmp(argv[i], "-h"))
    {
      if (i==argc-1) 
      {
        fprintf(stderr,"-h needs an argument\n");
        continue;
      }
      cmd_arg_hscale= strtof(argv[i+1], NULL);
      i++;
      continue;
    }
    fprintf(stderr,"unrecognized option %s\n", argv[i]);
  }
}

static void get_pixgrid()
{
  int bpl, shown;
  unsigned char *buffer;

  gx.pixgrid_attribs(Wplane, &shown, &disp.width, &disp.height);
  gfx_pixgrid_buffer(Wplane, &bpl, NULL, NULL, &buffer);

  disp.buffer= (void*) buffer;
  disp.wpl= bpl / 4;
  // ASSERT( bpp == 4 ); FIXME
}

void init_ro()
{
  ro_server= socket_server( 1605, 1 );
  gfx_monitor_file( ro_server, FD_SERVER, 1, 0, 0);
}


#define CURSORw  11
#define CURSORh  15
static const char CURSORd[]= {
0,0,0,2,2,2,2,2,2,2,2,
0,1,0,0,2,2,2,2,2,2,2,
0,1,1,0,0,2,2,2,2,2,2,
0,1,1,1,0,0,2,2,2,2,2,
0,1,1,1,1,0,0,2,2,2,2,
0,1,1,1,1,1,0,0,2,2,2,
0,1,1,1,1,1,1,0,0,2,2,
0,1,1,1,1,1,1,1,0,0,2,
0,1,1,1,1,1,1,1,1,1,0,
0,1,1,1,1,1,1,1,1,1,1,
0,1,1,1,1,0,0,0,0,0,0,
0,1,1,1,0,0,2,2,2,2,2,
0,1,1,0,0,2,2,2,2,2,2,
0,1,0,0,2,2,2,2,2,2,2,
0,0,0,2,2,2,2,2,2,2,2,
};
static uint32_t CURSORb[CURSORw*CURSORh];

static void update_cursor(int Ux0, int Uy0, int Uw, int Uh)
{
  int Ux1, Uy1, Cx0, Cx1, Cy0, Cy1;
  int x, y, uw, uh, ox, oy, dO, sO, wpl;
  uint32_t dP, *buffer; uint8_t sP;

  Cx0= CURSORx;           Ux1= Ux0 + Uw;     if (Cx0>=Ux1) return ;
  Cx1= Cx0 + CURSORw;                        if (Cx1<=Ux0) return ;
  if (Cx1>Ux1) { Cx1= Ux1; }                 if (Cx0<Ux0) { Cx0= Ux0; }

  Cy0= CURSORy;           Uy1= Uy0 + Uh;     if (Cy0>=Uy1) return ;
  Cy1= Cy0+CURSORh;                          if (Cy1<=Uy0) return ;
  if (Cy1>Uy1) { Cy1= Uy1; }                 if (Cy0<Uy0) { Cy0= Uy0; }

  ox= Cx0-CURSORx;                           oy= Cy0-CURSORy;
  uw= Cx1-Cx0;                               uh= Cy1-Cy0;

  wpl= disp.wpl;
  buffer= disp.buffer;

  for(y=0;y<uh;y++)
   for(x=0;x<uw;x++)
   {
     dO= (y+Cy0)*wpl + (x+Cx0);
     sO= (y+oy)*CURSORw+(x+ox);
     dP= buffer[dO];
     sP= CURSORd[sO];
     CURSORb[sO]= dP;
     buffer[dO]= sP == 2 ? dP : (sP ? 0xFFffFF : 0);
   }
}

static void draw_cursor()
{
  int Ux0, Uy0, Uw, Uh;
  int Ux1, Uy1, Cx0, Cx1, Cy0, Cy1;
  int x, y, uw, uh, ox, oy, dO, sO, wpl;
  uint32_t dP, *buffer; uint8_t sP;

  Ux0= 0;                                    Uw= disp.width;  
  Uy0= 0;                                    Uh= disp.height;

  Cx0= CURSORx;           Ux1= Ux0 + Uw;     if (Cx0>=Ux1) return ;
  Cx1= Cx0 + CURSORw;                        if (Cx1<=Ux0) return ;
  if (Cx1>Ux1) { Cx1= Ux1; }                 if (Cx0<Ux0) { Cx0= Ux0; }

  Cy0= CURSORy;           Uy1= Uy0 + Uh;     if (Cy0>=Uy1) return ;
  Cy1= Cy0+CURSORh;                          if (Cy1<=Uy0) return ;
  if (Cy1>Uy1) { Cy1= Uy1; }                 if (Cy0<Uy0) { Cy0= Uy0; }

  ox= Cx0-CURSORx;                           oy= Cy0-CURSORy;
  uw= Cx1-Cx0;                               uh= Cy1-Cy0;

  wpl= disp.wpl;
  buffer= disp.buffer;

  for(y=0;y<uh;y++)
   for(x=0;x<uw;x++)
   {
     dO= (y+Cy0)*wpl + (x+Cx0);
     sO= (y+oy)*CURSORw+(x+ox);
     dP= buffer[dO];
     sP= CURSORd[sO];
     CURSORb[sO]= dP;
     buffer[dO]= sP == 2 ? dP : (sP ? 0xFFffFF : 0);
   }
  gfx_pixgrid_update(Wplane, Cx0, Cy0, uw, uh);
}

static void undraw_cursor()
{
  int Ux0, Uy0, Uw, Uh;
  int Ux1, Uy1, Cx0, Cx1, Cy0, Cy1;
  int x, y, uw, uh, ox, oy, dO, sO, wpl;
  uint32_t *buffer; 

  Ux0= 0;                                    Uw= disp.width;  
  Uy0= 0;                                    Uh= disp.height;

  Cx0= CURSORx;           Ux1= Ux0 + Uw;     if (Cx0>=Ux1) return ;
  Cx1= Cx0 + CURSORw;                        if (Cx1<=Ux0) return ;
  if (Cx1>Ux1) { Cx1= Ux1; }                 if (Cx0<Ux0) { Cx0= Ux0; }

  Cy0= CURSORy;           Uy1= Uy0 + Uh;     if (Cy0>=Uy1) return ;
  Cy1= Cy0+CURSORh;                          if (Cy1<=Uy0) return ;
  if (Cy1>Uy1) { Cy1= Uy1; }                 if (Cy0<Uy0) { Cy0= Uy0; }

  ox= Cx0-CURSORx;                           oy= Cy0-CURSORy;
  uw= Cx1-Cx0;                               uh= Cy1-Cy0;

  wpl= disp.wpl;
  buffer= disp.buffer;

  for(y=0;y<uh;y++)
   for(x=0;x<uw;x++)
   {
     dO= (y+Cy0)*wpl + (x+Cx0);
     sO= (y+oy)*CURSORw+(x+ox);
     buffer[dO]= CURSORb[sO]; 
   }
  gfx_pixgrid_update(Wplane, Cx0, Cy0, uw, uh);
}

static int imin2(int a, int b) { return a<b ? a:b; }
static int imax2(int a, int b) { return a>b ? a:b; }

static void move_cursor(int x, int y)
{
  undraw_cursor();
  x= imin2(disp.width-1,imax2(0, x));   y= imin2(disp.height-1, imax2(0, y));
  CURSORx= x;                           CURSORy= y;
  draw_cursor();
}

static void print_ifad()
{
  char **ifad;
  int i;
  ifad= ifad_get();
  for(i=0;ifad[i];i++)
  {
      screen_print_str(tabs[0].screen, (const uint8_t*) ifad[i],
                                        strlen(ifad[i]));
      screen_print_str(tabs[0].screen, (const uint8_t*) "\r\n", 2); 
  } 
  ifad_free(ifad);
}


int main(int argc,char **argv)
{
  parse_args(argc, argv);



  if (getenv("MUTERM_PRINT_UESC")) matrix_debug(1);

  if (gfx_preinit(NULL)) return 1;

  init_ro();

  init_font(cmd_arg_fsize, cmd_arg_wextra, cmd_arg_hscale);
  init_colors();
  init_tabs();

  gfx_set_window_size(fontW*wd_cols, fontH*(wd_rows+1));
  gfx_init(NULL);

  gx.screen(Smain);
  get_pixgrid();
  
  draw_bar();
  activate(0);
  print_ifad();

  event_loop();

  gfx_close();
  return 0;
}
