/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include "udec.h"
#include "buffer.h"
#include "matrix.h"
#define UTIL_IMAX2
#define UTIL_IMIN2
#define UTIL_UMIN2
#define UTIL_UMAX2
#include "util.h"

#define ESC  '\033'

#define TXTATTR_ALT  1

#define DX printf

 
/*┌──────────────────────────────────────────────────────────────────────────┐
  │                             BASIC USAGE                                  │
  └──────────────────────────────────────────────────────────────────────────┘
    This matrix "library" is pretty independent from the rest of the program.
    Therefore, it makes sense to isolate it as much as possible and have it
    be used in other contexts.

    The main object is a screen_t. This object doesn't know how to display
    characters on the screen, it simply maintains a matrix of cells. The main
    program prints some strings to the matrix using screen_print_str(). The
    matrix interprets any escape codes in there and stores printable
    characters at the appropriate positions.

    The matrix then stores a list of cells to be updated. The main program
    calls screen_updates() when it's ready to update the display. The called
    function then calls set_cursor() and display_glyph() to update the state
    of the caller. Finally, it makes calls to update_area(). Separating 
    display_glyph() and update_area() in this way makes it possible for the
    main program to buffer updates and then apply them all at once during
    update_area(). This way glyphs don't need to go thru the display stack
    one by one.

  ┌──────────────────────────────────────────────────────────────────────────┐
  │                           THE TWO BUFFERS                                │
  └──────────────────────────────────────────────────────────────────────────┘

 One is the primary buffer, which is typically used for executing commands 
 etc. The other is the alternate buffer.  This buffer is usually employed
 for programs which do full-screen operations, such as editors, text-ui
 applications etc.

  ┌──────────────────────────────────────────────────────────────────────────┐
  │                            SCROLLBACK BUFFER                             │
  └──────────────────────────────────────────────────────────────────────────┘

 The primary buffer has a scrollback buffer. Lines that have been scrolled
 up beyond the first row are stored here. These can be reviewed later by
 using Ctrl-Pageup/Ctrl-Pagedown or mouse scrollwheel.

 The scrollback buffer will consist of a linked list of nodes which
 contain the following:

 buffer {
   size_t size, len;
   int seqno;
   int ncols;
   uint32_t *data;
   struct buffer *prev, *next;
 };

 Data in the screen will be held by a special buffer which is always full
 (size==len==ncols*nrows). This buffer points to the primary matrix. 
   
 Then, a position would be:

 position { struct buffer *buffer; size_t index; }

 The seqno field increases from top to bottom, i.e. from older content to
 newer content. It provides an easy way of comparing two positions.

 The ncols field holds the number of columns when the buffer was created.
 Recording this allows me to retain the current scrollback buffer when the
 window is resized. While viewing contents in a scrollback node, I just
 cut-off or pad the content if the current number of columns is different.

 When the user displays a scrolled buffer, the matrix itself isn't changed.
 Instead, the system simply updates the display from the scrollback buffer.
 The best place to do this is in the screen_updates() function. 

 While viewing the scrollback buffer, there are only two types of updates
 possible. In the first type, we're displaying the contents for the first
 time, due to a page_prev or page_next command. This type will also cover
 window resize events. The second update type happens when the selection
 changes and the currently viewed buffer contains part of the selection.
 In the first part, we simply paint everything and then store the top 
 position in the screen struct. To handle the latter, we find the position
 of the selection change and update everything stuff from there. Updates
 should be done in complete rows. This is the easiest way possible.

  ┌──────────────────────────────────────────────────────────────────────────┐
  │                             CELL ATTRIBUTES                              │
  └──────────────────────────────────────────────────────────────────────────┘
 These are:

  bold, faint, italic, underline, double underline, curly underline,
  blink, inverse, invisible, strikethrough

 All other attributes are related to color, which I don't do.
 I will use a reduced set of attributes, mapped as follows:

  bold      <= bold, faint
  italic    <= italic
  underline <= [curly | double] underline, strikethrough
  inverse   <= inverse
  invisible <= invisible
  blink     <= no fucking way

 Italic will be implemented as a different color because it's hard to 
 find a truly monospaced italic font which would play well with the 
 main font. I will initially implement the underline attribute as a
 different color as well. Later on, I can render the underline when the
 rest of the system is working properly.

  ┌──────────────────────────────────────────────────────────────────────────┐
  │                           GRAPHICS CHARACTERS                            │
  └──────────────────────────────────────────────────────────────────────────┘

   These characters are printed by switching to the special graphics character
   set. There is a list of these in 

       https://vt100.net/docs/vt100-ug/chapter3.html

   Look at Table 3-9. In order to offload the matrix user from interpreting
   these things, I simply convert the special characters to their Unicode
   counterparts. This way, when you copy/paste stuff that include these
   characters, they look OK in other contexts. The array "altset" holds the
   mapping between regular characters and special graphics symbols.

   There is a much more informative table at:

     https://wikimili.com/en/DEC_Special_Graphics

  ┌──────────────────────────────────────────────────────────────────────────┐
  │                                SELECTIONS                                │
  └──────────────────────────────────────────────────────────────────────────┘

    For the two matrices, the selection means different things. For the 
    primary matrix, all text that has been printed is available for 
    copy/paste. This means that the selection is a range defined by two 
    scrollback buffer positions, which may reside in the last buffer (the
    primary matrix).

    When the alternate matrix is active, then the only text available is
    what is shown on the screen. Therefore, the selection is defined by 
    two (row,column) tuples.

 ***********************************************************************/


static uint32_t altset[256];

typedef struct { int start, end; } upd_t;

#define  SB_ALLOC_SIZE 10

typedef struct
{
  int cursor_row, cursor_col;
  uint8_t m_corigin;
  uint8_t pw;
  uint32_t SGR;
} saved_cursor_t;

struct cbuf
{
  struct cbuf *prev, *next;
  int seqno;
  int ncols;
  size_t coff;
  size_t size, len;
  uint32_t *data;
};

typedef struct { struct cbuf *buf; size_t index; } pos_t;

#define SEL_INVAL   0
#define SEL_MAIN    1
#define SEL_ALT     2

typedef struct
{
  int type; 
  pos_t mpos;
  int row, col;
} selpos_t;

struct screen
{
  int out_fd; // for reporting things
  int    cursor_col, cursor_row;

  int nrows, ncols; 
  uint32_t *primary_matrix;
  uint32_t *alternate_matrix;
  saved_cursor_t *saved_cursor;
  uint32_t *matrix;
  uint8_t  *tabstop;

  int matrix_sel;  // 0 for normal, 1 for alternate

  /**** FLAGS ****/
  uint8_t pw; // pending wrap

  /**** MODES ****/
  uint8_t m_margins; // enable left,right margin command
  uint8_t m_corigin; // cursor origin at scroll top-left
  uint8_t m_chidden; 
  uint8_t m_insert, m_linefeed;
  uint8_t m_wrap, m_rewrap;
  uint8_t m_bs, m_del; // bs_sends_del delete_sends_del

  saved_cursor_t  primary_saved_cursor, alternate_saved_cursor;

  int decoding;
  utf8decoder_t decoder;

  uint32_t SGR; // text attributes such as bold etc.
  uint32_t TXTFLAGS; // such as TXTATTR_ALT etc.

          // currently parsing an escape sequence
  uint8_t escape_mode;

  uint8_t active; // if not set, changes will not be reflected on the screen
                  // it's a background matrix.

  // The following define the scrolling region. 
  int scroll_top, scroll_bot;
  int scroll_left, scroll_right;

  int rminupd, rmaxupd; upd_t *upd; int updates;

  int eslen;
  uint32_t esbuf[201];
  int esargs[100];
  int esatop;
  int estate;

#ifdef MUDEBUG
  // debugging buffer. I collect the escape sequence characters here
  // and then print it if the sequence isn't handled.
  // the same sequence is also collected in esbuf, but some chars
  // might be missing because I can omit stuff and just change
  // state in some cases.
  uint32_t esdbuf[201];
  int esdlen;
#endif

  /*** SCROLLBACK ***/
  struct cbuf *first, *last;
  uint8_t vs; // flag, are we viewing a scrollback buffer
  int     vsrow;  // first row to update in this mode
  int     vsnrow; // number of rows to update
  pos_t   updpos; // vsrow corresponds to this position
  pos_t   toppos;

  /*** SELECTION ***/
  int     sel_state; // 0 1 2 as described below
  int     sel_type;  // Which matrix does the selection refer to
  uint32_t sel_begin, sel_end; 
  uint32_t sel_row, sel_col;
};

#define MATRIX(S,r,c)  S->matrix[r*S->ncols+c]
#define ATTCHR(A,C)    (((C)&CELL_CHRMASK) | (A))

static int matrix_dbg;

static uint32_t cell_space;
static uint32_t cell_newline;

static void screen_updates_scrollback(screen_t *S);


static void init_altset()
{
  int i;
  static const uint32_t spec[]= {
                                                            0x00A0,
    0x25C6, 0x2592, 0x2409, 0x240C, 0x240D, 0x240A, 0x00B0, 0x00B1,
    0x2424, 0x240B, 0x2518, 0x2510, 0x250C, 0x2514, 0x253C, 0x23BA,
    0x23BB, 0x2500, 0x23BC, 0x23BD, 0x251C, 0x2524, 0x2534, 0x252C,
    0x2502, 0x2264, 0x2265, 0x03C0, 0x2260, 0x00A3, 0x00B7, 0 };
  if (altset['a']) return ;
  for(i=0;i<256;i++) altset[i]= i;
  for(i=0;spec[i];i++) altset[0x5F + i]= spec[i];
}

static void update_cell(screen_t *S, int row, int col)
{
  if (col<0 || col>=S->ncols || row<0 || row>=S->nrows) return ;
  S->updates= 1;
  S->upd[row].start= imin2(S->upd[row].start, col);
  S->upd[row].end= imax2(S->upd[row].end, col);
  S->rminupd= imin2(S->rminupd, row);
  S->rmaxupd= imax2(S->rmaxupd, row);
}

static void update_row(screen_t *S, int row, int cs, int ce)
{
  ce= imin2(ce, S->ncols-1);
  cs= imax2(cs, 0);
  if (cs<=ce)
  {
    S->updates= 1;
    S->upd[row].start= imin2(S->upd[row].start, cs);
    S->upd[row].end= imax2(S->upd[row].end, ce);
    S->rminupd= imin2(S->rminupd, row);
    S->rmaxupd= imax2(S->rmaxupd, row);
  }
}

static void update_rows(screen_t *S, int rs, int re, int cs, int ce)
{
  int r;
  rs= imax2(rs,0); re= imin2(re, S->nrows-1);
  cs= imax2(cs,0); ce= imin2(ce, S->ncols-1); 
  if (rs<=re && cs<=ce)
  {
    for(r=rs;r<=re;r++)
    {
      S->upd[r].start= imin2(S->upd[r].start, cs);
      S->upd[r].end= imax2(S->upd[r].end, ce);
    }
    S->rminupd= imin2(S->rminupd, rs);
    S->rmaxupd= imax2(S->rmaxupd, re);
    S->updates= 1;
  }
}

static void update_all(screen_t *S)
{
  int r;
  for(r=0;r<S->nrows;r++)
  {
    S->upd[r].start= 0;
    S->upd[r].end= S->ncols-1;
  }
  S->updates= 1;
  S->rminupd= 0;
  S->rmaxupd= S->nrows-1;
}


static void screen_updates_normal(screen_t *S)
{
  int cs, ce, r, e, z, c;
  uint32_t B,E, tp;
  if (!S->updates) return ;
  if (S->sel_type!=S->matrix_sel+1) 
  {
    set_sel(-1,0,0,0);
    goto draw;
  }

  B= umin2(S->sel_begin, S->sel_end);
  E= umax2(S->sel_begin, S->sel_end);
  if (S->matrix_sel==0)
  {
    tp= S->last->coff;
    if (E<tp) { set_sel(-1,0,0,0); goto draw; }
    B= umax2(B, tp);
    B-= tp;
    E-= tp;
  }
  set_sel(B/S->ncols, B%S->ncols, E/S->ncols, E%S->ncols);

draw:
  for(r=S->rminupd;r<=S->rmaxupd;r++)
   if (S->upd[r].start!=S->ncols)
   {
     cs= S->upd[r].start; ce= S->upd[r].end;
     for(e=r;e<=S->rmaxupd;e++)
      if (S->upd[e].start!=cs || S->upd[e].end!=ce) break;
      else { S->upd[e].start= S->ncols; S->upd[e].end= -1; }

     for(z=r;z<e;z++)
      for(c=cs;c<=ce;c++)
       display_glyph(z, c, MATRIX(S, z,c));
     update_area(cs, r, ce-cs+1, e-r);
     r= e-1;
   }
  S->updates= 0;
  S->rminupd= S->nrows-1;
  S->rmaxupd= 0;
  display_modified();
}

void screen_updates(screen_t *S)
{
  if (S->vs) 
     screen_updates_scrollback(S);
  else
     screen_updates_normal(S);
}

static void cursor(screen_t *S, int row, int col)
{
  row= imax2(0, imin2(S->nrows-1,row));
  col= imax2(0, imin2(S->ncols-1,col));
  if (S->cursor_row!=row || S->cursor_col!=col)
  {
    if (S->active && !S->m_chidden)
    {
      update_cell(S, S->cursor_row, S->cursor_col);
      update_cell(S, row, col);
      set_cursor(row, col);
    }
    S->cursor_row= row;
    S->cursor_col= col;
  }
}

static int cursor_in_sr(screen_t *S)
{
  return S->cursor_row>= S->scroll_top &&
         S->cursor_row<= S->scroll_bot &&
         S->cursor_col>= S->scroll_left &&
         S->cursor_col<= S->scroll_right;
}

static void save_cursor(screen_t *S)
{
  saved_cursor_t *C; 

  C= S->saved_cursor;
  C->m_corigin= S->m_corigin;
  C->SGR= S->SGR;
  C->pw= S->pw;

  if (C->m_corigin)
  {
    C->cursor_row= imax2(0, S->cursor_row-S->scroll_top);
    C->cursor_col= imax2(0, S->cursor_col-S->scroll_left);
  }
  else
  {
    C->cursor_row= S->cursor_row;
    C->cursor_col= S->cursor_col;
  }
}

static void restore_cursor(screen_t *S)
{
  saved_cursor_t *C; 
  int r, c;

  C= S->saved_cursor;
  S->m_corigin= C->m_corigin;
  S->SGR= C->SGR;
  S->pw= C->pw;
  
  if (S->m_corigin)
  {
    r= imin2(S->nrows-1, C->cursor_row+S->scroll_top);
    c= imin2(S->ncols-1, C->cursor_col+S->scroll_left);
  }
  else
  {
    r= C->cursor_row;
    c= C->cursor_col;
  }
  cursor(S, r,c);
}

static void tabstop_init(screen_t *S)
{
  int c;
  memset(S->tabstop, 0, sizeof(S->tabstop[0])*S->ncols);
  for(c=0;c<S->ncols;c+=8) S->tabstop[c]= 1;
}

static void tab(screen_t *S, int count)
{
  int i,C;
  C= S->cursor_col;
  while(count--)
  {
    for(i=C+1;i<S->ncols-1;i++) if (S->tabstop[i]) break;
    C= i;
  }
  cursor(S, S->cursor_row, C);
}

static void switch_to_alternate(screen_t *S)
{
  if (S->matrix_sel) return ;
  S->matrix= S->alternate_matrix;
  S->saved_cursor= &S->alternate_saved_cursor;
  S->matrix_sel= 1;
  update_all(S);
}

static void switch_to_primary(screen_t *S)
{
  if (!S->matrix_sel) return ;
  S->matrix= S->primary_matrix;
  S->saved_cursor= &S->primary_saved_cursor;
  S->matrix_sel= 0;
  update_all(S);
}

static void copy_row(screen_t *S, int margins, int dst, int src)
{
  int c, cs, ce;
  if (margins) { cs= S->scroll_left; ce= S->scroll_right; }
          else { cs= 0; ce= S->ncols-1; }
  for(c=cs;c<=ce;c++) MATRIX(S, dst, c)= MATRIX(S, src, c);
  update_row(S, dst, cs, ce);
}

static void erase_horz(screen_t *S, int r, int cs, int ce)
{
  int c;
  cs= imax2(cs,0);
  ce= imin2(ce,S->ncols-1);
  for(c=cs;c<=ce;c++) MATRIX(S, r, c)= cell_space;
  update_row(S, r, cs, ce);
}

static void erase_rows(screen_t *S, int margins, int rs, int re)
{
  int c, cs, ce, r;
  if (margins) { cs= S->scroll_left; ce= S->scroll_right; }
          else { cs= 0; ce= S->ncols-1; }
  for(r=rs;r<=re;r++)
  for(c=cs;c<=ce;c++)
    MATRIX(S, r, c)= cell_space;
  update_rows(S, rs, re, cs, ce);
}

static void print_chrs_simple(screen_t *S, const uint8_t *str, int len)
{
  int i,r,c, do_alt;
  uint32_t SGR;

  r= S->cursor_row;
  c= S->cursor_col;
  do_alt= (S->TXTFLAGS & TXTATTR_ALT) ? 1 : 0;
  SGR= (S->SGR | S->TXTFLAGS) & ~TXTATTR_ALT;

  if (do_alt)
     for(i=0;i<len;i++) MATRIX(S, r, c+i)= ATTCHR(SGR, altset[str[i]]);
  else
     for(i=0;i<len;i++) MATRIX(S, r, c+i)= ATTCHR(SGR, str[i]);

  update_row(S, r, c, c+i);
  cursor(S, r, c+i);
  return ;
}

static void fill_matrix(screen_t *S, uint32_t *M, uint8_t c)
{
  uint32_t chr;
  int i,T;
  T= S->nrows*S->ncols;
  chr= S->SGR | c;
  for(i=0;i<T;i++) M[i]= chr;
}

static void copy_rows_to_scrollback(screen_t *S,int n)
{
  struct cbuf *B;
  uint32_t *src;
  size_t cp;

  src= S->last->data;
  n*= S->ncols;
again:
  if (!n) return ;
  B= S->last->prev;
  if (!B || B->size==B->len || B->ncols!=S->ncols)
  {
    B= calloc(1,sizeof(*B));
    B->size= SB_ALLOC_SIZE*S->nrows*S->ncols;
    B->data= malloc(sizeof(B->data[0])*B->size);
    B->ncols= S->ncols;
    B->prev= S->last->prev;
    B->next= S->last;
    if (S->last->prev) S->last->prev->next= B;
    S->last->prev= B;
    B->seqno= S->last->seqno++;
    B->coff= S->last->coff;
  }
  cp= B->size - B->len;
  if (cp>n) cp= n;
  memcpy(B->data+B->len, src, cp*sizeof(uint32_t));
  B->len+= cp;
  S->last->coff += cp;
  src+= cp;
  n-= cp;
  goto again;
}


/** ##C : Commands **/

static int tCR(screen_t *S)
{
  int cc;
  S->pw= 0;
  if (S->m_corigin)
  {
    cc= S->scroll_left;
  }
  else
  {
     if (S->cursor_col>=S->scroll_left)
        cc= S->scroll_left;
     else
        cc= 0;
  }
  cursor(S, S->cursor_row, cc);
  return 1;
}

/*** Commands Part 1: Movement ***/

static int tCUU(screen_t *S, int n)
{
  int max_m;
  int R;
  S->pw= 0;
  R= S->cursor_row;
  if (R>=S->scroll_top) max_m= R-S->scroll_top;
                   else max_m= R;
  n= imin2(max_m, imax2(n,1));
  cursor(S, R-n, S->cursor_col);
  return 1;
}

static int tCUD(screen_t *S,int n)
{
  int max_m;
  int R;
  S->pw= 0;
  R= S->cursor_row;
  if (R<=S->scroll_bot) max_m= S->scroll_bot-R;
                   else max_m= S->nrows-1-R;
  n= imin2(max_m, imax2(n,1));
  cursor(S, R+n, S->cursor_col);
  return 1;
}

static int tCUF(screen_t *S,int n)
{
  int max_m;
  int C;
  S->pw= 0;
  C= S->cursor_col;
  if (C<=S->scroll_right) max_m= S->scroll_right-C;
                     else max_m= S->ncols-1-C;
  n= imin2(max_m, imax2(n,1));
  cursor(S, S->cursor_row, C+n);
  return 1;
}

static int tCUB(screen_t *S,int n)
{
  int max_m;
  int C, R, edge;

  C= S->cursor_col;
  if (S->m_wrap && S->m_rewrap) goto its_complicated;

  S->pw= 0;
  if (C>=S->scroll_left) max_m= C-S->scroll_left;
                    else max_m= C;
  n= imin2(max_m, imax2(n,1));
  cursor(S, S->cursor_row, C-n);
  return 1;
its_complicated:
  n= imax2(n,1);
  if (S->pw) { S->pw= 0; n--; }
  R= S->cursor_row;
  while(n)
  {
    if (C>=S->scroll_left) edge= S->scroll_left;
                      else edge= 0;
    if (C-n>=edge) { C-=n; break; }
    if (C==edge)
    {
      n--;
      R= (R+S->nrows-1)%S->nrows;
      C= S->scroll_right;
    }
    else
    {
      n-= C-edge;
      C= edge;
    }
  }
  cursor(S, R, C);
  return 1;
}

static int tCUP(screen_t *S,int dr,int dc)
{
  S->pw= 0;
  dr= imax2(dr,1)-1;
  dc= imax2(dc,1)-1;
  if (S->m_corigin)
  {
    dr= imin2(S->scroll_bot, S->scroll_top+dr);
    dc= imin2(S->scroll_right, S->scroll_left+dc);
  }
  else
  {
    dr= imin2(dr, S->nrows-1);
    dc= imin2(dc, S->ncols-1);
  }
  cursor(S, dr, dc);
  return 1;
}

static int tVPA(screen_t *S,int dr)
{
  int dc;
  S->pw= 0;
  dc= S->cursor_col;
  dr= imax2(dr,1)-1;
  if (S->m_corigin)
  {
    dr= imin2(S->scroll_bot, S->scroll_top+dr);
    dc= imax2(S->scroll_left, imin2(S->scroll_right,dc));
  }
  else
  {
    dr= imax2(0, imin2(S->nrows-1, dr));
  }
  cursor(S, dr, dc);
  return 1;
}

static int tHPA(screen_t *S,int dc)
{
  int dr;
  S->pw= 0;
  dr= S->cursor_row;
  dc= imax2(dc,1)-1;
  if (S->m_corigin)
  {
    dr= imax2(S->scroll_bot, imin2(S->scroll_top, dr));
    dc= imin2(S->scroll_right, S->scroll_left+dc); 
  }
  else
  {
    dc= imax2(S->scroll_left, imin2(S->scroll_right, dc));
  }
  cursor(S, dr, dc);
  return 1;
}

static int tCNL(screen_t *S, int n)
{
  n= imax2(1,n);
  tCUD(S,n);
  tCR(S);
  return 1;
}

static int tCPL(screen_t *S, int n)
{
  n= imax2(1,n);
  tCUU(S,n);
  tCR(S);
  return 1;
}

static int tHPR(screen_t *S, int n)
{
  int R,C;
  n= imax2(1,n);
  S->pw= 0;
  R= S->cursor_row;
  C= S->cursor_col + n;
  if (S->m_corigin) R= imax2(S->scroll_top, imin2(R, S->scroll_bot));
  C= imin2(C, S->scroll_right);
  cursor(S, R, C);
  return 1;
}

static int tVPR(screen_t *S, int n)
{
  int R,C;
  S->pw= 0;
  C= S->cursor_col;
  R= S->cursor_row + n;
  if (S->m_corigin) C= imax2(S->scroll_left, imin2(C, S->scroll_right));
  R= imin2(R, S->scroll_bot);
  cursor(S, R, C);
  return 1;
}

static int tCHT(screen_t *S, int n)
{
  tab(S, imax2(n,1));
  return 1;
}

static int tCBT(screen_t *S, int n)
{
  int C,i;
  n= imax2(n,1);
  C= S->cursor_col;
  while(C>0 && n--)
  {
    for(i=C-1;i>=0;i--) if (S->tabstop[i]) break;
    C= i;
  }
  C= imax2(C,0);
  if (S->m_corigin) C= imax2(C, S->scroll_left);
  cursor(S, S->cursor_row, C);
  return 1;
}

/*** Commands Part 2: Scrolling ***/

static int tSU(screen_t *S, int n)
{
  int r;
  n= imin2(S->scroll_bot-S->scroll_top+1,imax2(n,1));

  if (!S->matrix_sel && S->scroll_top==0 && S->scroll_left==0 &&
       S->scroll_right==S->ncols-1)
    copy_rows_to_scrollback(S, n);

  for(r=S->scroll_top;r+n<=S->scroll_bot;r++) copy_row(S, 1, r, r+n);
  erase_rows(S, 1, r, S->scroll_bot);
  return 1;
}

static int tSD(screen_t *S, int n)
{
  int r;
  n= imax2(n,1);
  for(r=S->scroll_bot;r-n>=S->scroll_top;r--) copy_row(S, 1, r, r-n);
  erase_rows(S, 1, S->scroll_top, r);
  return 1;
}

static int tIND(screen_t *S)
{ 
  int cr;
  cr= S->cursor_row;
  S->pw= 0;
  if (cursor_in_sr(S) && cr==S->scroll_bot) tSU(S, 1);
                                       else cursor(S, cr+1, S->cursor_col);
  return 1;
}

static int tRI(screen_t *S)
{
  int cr;
  cr= S->cursor_row;
  S->pw= 0;
  if (cursor_in_sr(S) && cr==S->scroll_top) tSD(S, 1);
                                       else cursor(S, cr-1, S->cursor_col);
  return 1;
} 

static int tDECBI(screen_t *S)
{
  int cr, r,c;
  if (S->cursor_col!=S->scroll_left) return tCUB(S,1);
  cr= S->cursor_row;
  if (cr<S->scroll_top || cr>S->scroll_bot) return 1;
  for(r=S->scroll_top;r<=S->scroll_bot;r++)
  {
    for(c=S->scroll_right;c>S->scroll_left;c--)
      MATRIX(S, r, c)= MATRIX(S, r, c-1);
    MATRIX(S, r, c)= cell_space;
  }
  update_rows(S, S->scroll_top, S->scroll_bot, 
                 S->scroll_left, S->scroll_right);
  return 1;
}

static int tDECFI(screen_t *S)
{
  int cr, r,c;
  if (S->cursor_col!=S->scroll_right) return tCUB(S,1);
  cr= S->cursor_row;
  if (cr<S->scroll_top || cr>S->scroll_bot) return 1;
  for(r=S->scroll_top;r<=S->scroll_bot;r++)
  {
    for(c=S->scroll_left;c<S->scroll_right;c++)
      MATRIX(S, r, c)= MATRIX(S, r, c+1);
    MATRIX(S, r, c)= cell_space;
  }
  update_rows(S, S->scroll_top, S->scroll_bot, 
                 S->scroll_left, S->scroll_right);
  return 1;
}

/*** Commands Part 3: Insertion and Deletion ***/

static int tICH(screen_t *S,int n)
{
  int cc, cr, c, mr;
  S->pw= 0;
  if (!cursor_in_sr(S)) return 0;
  n= imax2(n,1);
  cr= S->cursor_row;
  cc= S->cursor_col;
  mr= S->scroll_right;
  for(c=mr;c-n>=cc;c--) MATRIX(S, cr, c)= MATRIX(S, cr, c-n);
  for(      ;c>=cc;c--) MATRIX(S, cr, c)= cell_space;
  update_row(S, cr, cc, mr);
  return 1;
}

static int tDCH(screen_t *S,int n) 
{
  int cc, cr, c,mr;
  S->pw= 0;
  if (!cursor_in_sr(S)) return 0;
  n= imax2(n,1);
  cr= S->cursor_row;
  cc= S->cursor_col;
  mr= S->scroll_right;
  for(c=cc;c+n<=mr;c++) MATRIX(S, cr, c)= MATRIX(S, cr, c+n);
  for(      ;c<=mr;c++) MATRIX(S, cr, c)= cell_space;
  update_row(S, cr, cc, mr);
  return 1;
}

static int tIL(screen_t *S,int n)
{ 
  int r;
  S->pw= 0;
  if (!cursor_in_sr(S)) return 0;
  n= imax2(1, n);
  for(r=S->scroll_bot;r-n>=S->cursor_row;r--) copy_row(S, 1, r, r-n);
  erase_rows(S, 1, S->cursor_row, r);
  cursor(S, S->cursor_row, S->scroll_left);
  return 1;
} 

static int tDL(screen_t *S,int n)
{
  int r;
  S->pw= 0;
  if (!cursor_in_sr(S)) return 0;
  n= imax2(1, n);
  for(r=S->cursor_row;r+n<=S->scroll_bot;r++) copy_row(S, 1, r, r+n);
  erase_rows(S, 1, r, S->scroll_bot);
  cursor(S, S->cursor_row, S->scroll_left);
  return 1;
} 

/*** Commands Part 4: Erasing ***/

static int tECH(screen_t *S,int n)
{
  S->pw= 0;
  erase_horz(S, S->cursor_row, S->cursor_col,
            imin2(S->ncols-1, S->cursor_col+imax2(n,1)-1));        
  return 1;
}

#define tDECSEL tEL
static int tEL(screen_t *S,int n)
{
  int r;
  r= S->cursor_row;
  switch(n)
  {
  case 3: if (S->pw) return 1; // fallthru
  case 0: erase_horz(S, r, S->cursor_col, S->ncols-1); break;
  case 1: erase_horz(S, r, 0, S->cursor_col); break;
  case 2: erase_horz(S, r, 0, S->ncols-1); break;
  default: return 0;
  }
  S->pw= 0;
  return 1;
}

#define tDECSED tED
static int tED(screen_t *S,int n) 
{
  switch(n) 
  {
  case 0: if (S->cursor_row==0 && S->cursor_col==0) break;
          erase_horz(S, S->cursor_row, S->cursor_col, S->ncols-1);
          erase_rows(S, 0, S->cursor_row+1, S->nrows-1);
          S->pw= 0;
          return 1;
  case 1: if (S->cursor_row==S->nrows-1 && S->cursor_col==S->ncols-1) break;
          erase_horz(S, S->cursor_row, 0, S->cursor_col);
          erase_rows(S, 0, 0, S->cursor_row-1);
          S->pw= 0;
          return 1;
  case 2: break;
  case 3: return 1; // ERASE SCROLLBACK BUFFER
  default: return 0;
  }
  erase_rows(S, 0, 0, S->nrows-1);
  S->pw= 0;
  return 1;
}

/*** Commands Part 5: Alternate Buffer & Save/Restore Cursor ***/

static int tDECSC(screen_t *S)
{
  save_cursor(S);
  return 1;
}

static int tDECRC(screen_t *S)
{
  restore_cursor(S);
  return 1;
}


/*** Commands Part 6: Settings ***/

static int tSGR(screen_t *S,int n)
{
 switch(n)
 {
 case 0: S->SGR= 0; break;
 case 1: S->SGR |= CELLATTR_BOLD; break;
 case 2: break;                          // faint
 case 3: break;                          // italic
 case 4: S->SGR |= CELLATTR_ULINE; break;
 case 5: S->SGR |= CELLATTR_BLINK; break; 
 case 7: S->SGR |= CELLATTR_REV; break;
 case 8: S->SGR |= CELLATTR_HIDDEN; break;
 case 22: S->SGR &= ~CELLATTR_BOLD; break;
 case 23: break;                        // italic
 case 24: S->SGR &= ~CELLATTR_ULINE;break;
 case 25: S->SGR &= ~CELLATTR_BLINK; break;
 case 27: S->SGR &= ~CELLATTR_REV; break;
 case 28: S->SGR &= ~CELLATTR_HIDDEN; break;
 }
 return 1;
}

static int tDECSTBM(screen_t *S,int T,int B)
{
  if (!B || B>S->nrows) B= S->nrows;
  if (!T) T= 1;

  if (T<B)
  {
    S->scroll_top= T-1;
    S->scroll_bot= B-1;
    tCUP(S, 1, 1);
  }
  else
  {
    S->scroll_top= 0;
    S->scroll_bot= S->nrows-1;
  }
  return 1;
}

static int tDECSLRM(screen_t *S, int left, int right)
{
  if (!S->m_margins)
  {
    // FIXME: save cursor, just do it like xterm does it.
    return 0;
  }

  if (!right || right>=S->ncols) right= S->ncols;
  if (left>=S->ncols) left= S->ncols;
  if (left>=right) return 0;
  S->scroll_left= left-1;
  S->scroll_right= right-1;
  tCUP(S,1,1);
  return 1;
}

static int tSM(screen_t *S,int qm,int n)
{
 if (!qm) goto noqm;
 switch(n)
 {
 case 6:     S->m_corigin= 1; tCUP(S, 1, 1); break;
 case 7:     S->m_wrap= 1; S->pw= 0; break; // FIXME: I put S->pw here..
 case 25:    S->m_chidden= 0;
             if (S->active) set_cursor(S->cursor_row, S->cursor_col);
             update_cell(S, S->cursor_row, S->cursor_col);
             break;
 case 45:    S->m_rewrap=1; break;
 case 47:
 case 1047:  if (S->matrix_sel) break;
             switch_to_alternate(S);
             break;
 case 1048:  save_cursor(S); break;
 case 1049:  if (S->matrix_sel) break;
             save_cursor(S);
             switch_to_alternate(S);
             fill_matrix(S, S->alternate_matrix, ' ');
             break;
 case 67:    S->m_bs= 1; break;
 case 69:    S->m_margins= 1; break;
 case 1037:  S->m_del= 1; break;
 default:    return 0;
 }
 return 1;
noqm:
 switch(n)
 {
 case 4:     S->m_insert=1; break;
 case 20:    S->m_linefeed=1; break;
 default:    return 0;
 }
 return 1;
}

static int tRM(screen_t *S,int qm, int n)
{
 if (!qm) goto noqm;
 switch(n)
 {
 case 6:     S->m_corigin= 0; break;
 case 7:     S->m_wrap= 0; S->pw= 0; break; // FIXME: I put S->pw here
 case 25:    S->m_chidden= 1;
             if (S->active) set_cursor(-1,-1);
             update_cell(S,S->cursor_row,S->cursor_col);
             break;
 case 45:    S->m_rewrap= 0; break;
 case 47:    if (!S->matrix_sel) break;
             switch_to_primary(S);
             break;
 case 1047:  if (!S->matrix_sel) break;
             switch_to_primary(S);
             fill_matrix(S, S->alternate_matrix, ' ');
             break;
 case 1048:  restore_cursor(S); break;
 case 1049:  if (!S->matrix_sel) break;
             switch_to_primary(S);
             restore_cursor(S);
             break;
 case 67:    S->m_bs= 0; break;
 case 69:    S->m_margins= 0;
             S->scroll_left= 0;
             S->scroll_right= S->ncols-1;
             break;
 case 1037:  S->m_del= 0; break;
 default:    return 0;
 }
 return 1;
noqm:
 switch(n)
 {
 case 4:     S->m_insert= 0; break;
 case 20:    S->m_linefeed= 0; break;
 default:    return 0;
 }
 return 1;
}

static int tHTS(screen_t *S)
{
  S->tabstop[S->cursor_col]= 1;
  return 1;
}

static int tTBC(screen_t *S, int n)
{
  switch(n)
  {
  case 0: S->tabstop[S->cursor_col]= 0; break;
  case 3: memset(S->tabstop, 0, sizeof(S->tabstop)*S->ncols); break;
  default: return 0;
  }
  return 1;
}

static int tCTC(screen_t *S, int n)
{
  switch(n)
  {
  case 0: tHTS(S); break;
  case 2: tTBC(S, 0); break;
  case 5: tTBC(S, 3); break;
  default: return 0;
  }
  return 1;
}

/*** Commands Part 7: Miscellaneous ***/

static int prt_posint(char *out, int val)
{
  char B[20];
  int L;
  sprintf(B,"%d",val);
  L= strlen(B);
  memcpy(out, B, L);
  return L; 
}

static int runOSC(screen_t *S)
{
  int i,V,R;
  size_t TL;
  uint8_t *T;
  char buf[200];

  R= 0;
  V= 0;
  for(i=0;i<S->eslen;i++)
  {
    if (S->esbuf[i]==';') break;
    if (S->esbuf[i]>='0' && S->esbuf[i]<='9')
       V= V*10+S->esbuf[i]-'0';
  }
  if (i==S->eslen) return 0;
  i++;
  T= malloc((S->eslen-i+1)*6);
  utf8_encode_string(S->esbuf+i,S->eslen-i,T,&TL);
  T[TL]= 0;
  // printf("OSC string: '%s'\n", T);
  switch(V)
  {
  case 0:
  case 2:
    set_title(T);
    R= 1;
    break;
  case 10: case 11: case 12: case 13:
  case 14: case 15: case 16: case 17:
  case 18: case 19:
    /* Normally, these are used for setting the dynamic colors.
     * The syntax is: OSC 10;R/G/B ST. Here, R, G and B are placeholders
     * for the RGB components of the corresponding color. 
     * However, since I'm not going to implement these things, I will
     * simply ignore them.
     *
     * Another syntax is:
     * OSC 10;?ST 
     * This should result in a string which restores the given
     * dynamic color to its original value. I will just print some
     * random colors here. It's fine.
    */
    R= 1;
    if (T[0]!='?') break;
    i= 0;
    buf[i++]= ESC; buf[i++]= ']'; i+= prt_posint(buf+i,V);
    buf[i++]= ';'; i+= prt_posint(buf+i, 20); buf[i++]= ESC; buf[i++]= '\\';
    write(S->out_fd, buf,i);
    break;
  }
  free(T);
  return R;
}


/*
   CSI Ps n  Device Status Report (DSR).
            Ps = 5  -> Status Report.  Result (``OK'') is
          CSI 0 n
            Ps = 6  -> Report Cursor Position (CPR) [row;column].
          Result is
          CSI r ; c R
 */
static int tDSR(screen_t *S,int n)
{
  char buf[200];
  int i;

  switch(n)
  {
  case 5:
    i= 0;
    buf[i++]= ESC;
    buf[i++]= '[';
    buf[i++]= '0'; 
    buf[i++]= 'n';
    write(S->out_fd, buf, n);
    break;
  case 6: 
    i= 0;
    buf[i++]= ESC;
    buf[i++]= '[';
    i+= prt_posint(buf+i, S->cursor_row);
    buf[i++]= ';';
    i+= prt_posint(buf+i, S->cursor_col);
    buf[i++]= 'R';
    write(S->out_fd, buf, n);
    break;
  }
  return 1;
}

static int tFULL_RESET(screen_t *S)
{
  tabstop_init(S);
  S->pw= 0;
  S->m_corigin= 0;
  S->SGR= 0;
  S->TXTFLAGS= 0;
  S->scroll_left= 0;
  S->scroll_right= S->ncols-1;
  S->scroll_top= 0;
  S->scroll_bot= S->nrows-1;
  S->m_insert= 0;
  S->m_wrap= 1;
  S->m_rewrap= 0;
  switch_to_primary(S);
  S->pw= 0;
  S->m_corigin= 0;
  S->SGR= 0;
  save_cursor(S);
  tED(S, 2);
  tCUP(S,1,1);
  update_all(S);
  return 1;
}

static int tDECALN(screen_t *S)
{
  int r,c;
  for(r=0;r<S->nrows;r++) for(c=0;c<S->ncols;c++)
    MATRIX(S,r,c)= 'E';
  S->scroll_left= 0;
  S->scroll_right= S->ncols-1;
  S->scroll_top= 0;
  S->scroll_bot= S->nrows-1;
  cursor(S,0,0);
  update_all(S);
  return 1;
}

#ifdef MUDEBUG
static int TERMDEBUG(screen_t *S)
{
  return 1;
}
#endif

static int tMEDIACONTROL(screen_t *S,int n)
{
  FILE *f;
  char *fn;
  int r,c;
  unsigned int chr;
  switch(n)
  {
  case 0: goto print_all;
  default: break;
  }
  return 0;
print_all:
  fn= getenv("MUTERMOUTFILE");
  if (!fn) fn= "muterm-out";
  f= fopen(fn, "wb");
  for(r=0;r<S->nrows;r++)
  {
    for(c=0;c<S->ncols;c++)
    {
      chr= MATRIX(S, r, c) & CELL_CHRMASK;
      if (chr>=' ' && chr<127) fputc(chr, f);
                          else fputc('.', f);
    }
    fputc('\n', f);
  }
  fclose(f);
  return 1;
}


/** ##P : Parser **/

static int loop1(screen_t *S,int (*f)(screen_t*,int), int def1)
{
  int i;
  int V = 1;
  for(i=0;i<=S->esatop;i++)
    V &= (*f)(S, S->esargs[i] ? S->esargs[i] : def1);
  return V;
}

static int loop1a(screen_t *S,int (*f)(screen_t*,int,int), int arg, int def1)
{
  int i;
  int V = 1;
  for(i=0;i<=S->esatop;i++)
    V &= (*f)(S, arg, S->esargs[i] ? S->esargs[i] : def1);
  return V;
}

static int simple1(screen_t *S,int (*f)(screen_t *,int),int def1)
{
  return (*f)(S,S->esargs[0] ? S->esargs[0] : def1);
}


static int simple2(screen_t *S,int (*f)(screen_t*,int,int),int def1,int def2)
{
  int a,b;
  a= S->esargs[0]; 
  a= a ? a : def1;
  if (S->esatop) b= S->esargs[1]; else b= 0;
  b= b ? b : def2;
  return (*f)(S,a,b);
}

static void escput(screen_t *S,uint32_t chr)
{
  if (S->eslen<199) S->esbuf[S->eslen++]= chr;
}

static void escdigit(screen_t *S,uint32_t chr)
{
   S->esargs[S->esatop]= S->esargs[S->esatop]*10 + chr - '0';
}

#ifdef MUDEBUG
static void esdput(screen_t *S,uint32_t chr)
{
  if (S->esdlen<199) S->esdbuf[S->esdlen++]= chr;
}

static void esdprint(screen_t *S,int H)
{
  int i;
  printf("%d: \\E",H);
  for(i=0;i<S->esdlen;i++)
     if (S->esdbuf[i]>' ' && S->esdbuf[i]<127)
       printf("%c",S->esdbuf[i]);
     else
       printf(" %02x ",S->esdbuf[i]);
  printf("\n");
}
#endif

static void escape_chr(screen_t *S,uint32_t chr)
{
  int H __attribute__((unused));
  H= 0;
#ifdef MUDEBUG
  esdput(S, chr);
#endif

  if (chr==0x18 || chr==0x1a) { S->escape_mode= 0; return ; }

#define STATES \
  STATE(INIT)  STATE(CSI) STATE(CSINORMAL) STATE(CSIQUES)\
  STATE(DCS) STATE(OSC) STATE(OSCcheckST) \
  STATE(DESIG0) \
  STATE(PM) STATE(APC) STATE(FINISTR) STATE(IGNORE1CHAR) \
  STATE(ESCHASH)

  enum { 
#define STATE(x)  s##x,
    STATES
#undef STATE
  };
#define JMP(x)  do { S->estate= s##x; return ; } while(0)

  switch(S->estate) {
#define STATE(x) case s##x: goto x;
STATES
#undef STATE
  }
  return ;
  // consumed ESC
INIT:
  switch(chr)
  {
  case '6': H= tDECBI(S); break;
  case '7': H= tDECSC(S); break;
  case '8': H= tDECRC(S); break;
  case '9': H= tDECFI(S); break;
  case 'D': H=tIND(S); break;
  case 'E': tCR(S); H= tIND(S); break;
  //case 'F': H=tCURSOR_LEFT_LOWER(); break;
  case 'H': H=tHTS(S); break;
  case 'M': H=tRI(S); break;
  case 'N': 
  case 'O': break;
  case 'P': JMP(DCS);
  case 'V': 
  case 'W': break;
  case 'X': // SOS: what's this?
  // case 'Z': H=tDECID(); break;
  case '[': S->eslen= 0; S->esatop= 0; S->esargs[0]= 0; JMP(CSI);
  case '\\': break; // lone string terminator, what are you doing here?
  case ']': S->eslen=0; JMP(OSC);
  case '^': JMP(PM);
  case '_': JMP(APC);
  case '#': JMP(ESCHASH);
  case ' ':
  case '%':
  case ')':
  case '*':
  case '+': JMP(IGNORE1CHAR);
  case '(': JMP(DESIG0);
  case 'c': H=tFULL_RESET(S); break;
#ifdef MUDEBUG
  case '0': H= TERMDEBUG(S); break; // this code normally doesn't exist.
#endif
  case '>':
  case '=':
  case '|':
  case '}': goto fini;
  
  }
IGNORE1CHAR:  
  goto fini;
OSC:
  if (chr=='\a') goto OSCdone;
  if (chr==ESC) JMP(OSCcheckST);
  if (chr==0x9c) goto OSCdone;
  escput(S,chr);
  return ;
OSCcheckST:
  if (chr=='\\') goto OSCdone;
  if (chr==ESC) return ;
  escput(S,chr);
  JMP(OSC);
OSCdone:
  H= runOSC(S);
  goto fini;
  
PM: while(0) ;
APC: while(0) ;  // fallthru
DCS:
  if (chr==ESC) JMP(FINISTR);
  return ;
FINISTR:
  if (chr=='\\') goto fini;
  if (chr==ESC) return ;
  JMP(DCS);

CSI:
  if (chr=='?') JMP(CSIQUES);
  // if (chr=='>') JMP(CSIGREA);
  S->estate= sCSINORMAL;
CSINORMAL:
  escput(S,chr);
  switch(chr)
  {
  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9': 
     escdigit(S,chr);
     return ;
  case ';':
     S->esatop++;
     S->esargs[S->esatop]= 0;
     return ;
  case 'A': H=simple1(S,tCUU, 1); break;
  case 'B': H=simple1(S,tCUD, 1); break;
  case 'C': H=simple1(S,tCUF, 1); break;
  case 'D': H=simple1(S,tCUB, 1); break;
  case 'E': H=simple1(S,tCNL,1);  break;
  case 'F': H=simple1(S,tCPL,1);  break;
  case 'G': H=simple1(S,tHPA,1);  break;
  case 'H': H=simple2(S,tCUP,1,1); break;
  case 'I': H=simple1(S,tCHT,1); break;
  case 'J': H=simple1(S,tED, 0); break;
  case 'K': H=simple1(S,tEL, 0); break;
  case 'L': H=simple1(S,tIL, 1); break;
  case 'M': H=simple1(S,tDL, 1); break;
  case 'P': H=simple1(S,tDCH,1); break;
  case 'S': H=simple1(S, tSU,1); break; 
  case 'T': H=simple1(S,tSD,1);  break;  
  case 'W': H=simple1(S,tCTC,0); break;
  case 'X': H=simple1(S,tECH,1); break;
  case 'Z': H=simple1(S,tCBT,1); break;
  case '@': H=simple1(S,tICH,1); break;
  case '`': H=simple1(S,tHPA,1); break;
  case 'a': H=simple1(S,tHPR,1); break;
  case 'd': H=simple1(S,tVPA, 1); break;
  case 'e': H=simple1(S,tVPR,1); break;
  case 'f': H=simple2(S,tCUP,1,1); break; // same as H 
  case 'g': H=simple1(S,tTBC,0); break;
  case 'h': H=loop1a(S,tSM,0,0); break;
  case 'l': H=loop1a(S,tRM,0,0); break;
  case 'm': H=loop1(S,tSGR, 0); break;
  case 'n': H=simple1(S,tDSR,1); break;
  case 'r': H=simple2(S,tDECSTBM, 1, S->nrows); break;
  case 'i': H=simple1(S,tMEDIACONTROL, 0); break;
  case 'j': H=simple1(S,tCUB,1); break;
  case 'k': H=simple1(S,tCUU,1); break;
  case 's': H=simple2(S,tDECSLRM, 1, 0); break;
  case 'u': H=tDECRC(S); break;
  }
  goto fini;
CSIQUES:
  escput(S, chr);
  switch(chr)
  {
  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9': 
     escdigit(S,chr);
     return ;
  case ';': S->esatop++; S->esargs[S->esatop]= 0; return ;
  case 'J': H=simple1(S,tDECSED,0); break;
  case 'h': H=loop1a(S,tSM,1,0); break;
  case 'l': H=loop1a(S,tRM,1,0); break;
  case 'K': H=simple1(S,tDECSEL, 0); break;
  case 'W': tabstop_init(S); H=1; break;
  }
fini:
  S->escape_mode= 0;
#ifdef MUDEBUG
  if (matrix_dbg) esdprint(S, H);
#endif
  return ;
DESIG0:
  switch(chr)
  {
  case '0': S->TXTFLAGS |= TXTATTR_ALT; H= 1; break;
  case 'B': S->TXTFLAGS &= ~TXTATTR_ALT; H= 1; break; 
  }
  goto fini;
ESCHASH:
  escput(S,chr);
  switch(chr)
  {
  case '8': H= tDECALN(S); break;
  }
  goto fini;
}


/**
  Documented control characters that are not implemented here are 
  as follows:
  0x05 ENQ : should print some string back to out_fd,
             currently xterm has a configurable string.
  0x07 BEL : I hate bells.
  0x0E SO    
  0x0F SI  : these deal with old-style character sets.
  0x18 CAN
  0x1A SUB : these abort the current escape sequence. they
             are normally handled in the command parser.
             if they appear outside an escape sequence,
             there is no effect.
 **/

static void control_chr(screen_t *S, uint32_t chr)
{

  switch(chr)
  {
/* BS  */  case 0x08: tCUB(S, 1); return ;
/* TAB */  case 0x09: tab(S,1);
                      return ;
/* LF  */  case 0x0a:
/* VT  */  case 0x0b:
/* FF  */  case 0x0c: tIND(S); if (S->m_linefeed) tCR(S); return ;
/* CR  */  case 0x0d: tCR(S); return ;
/* ESC */  case 0x1b: S->escape_mode= 1;
                      S->estate= 0;
                      S->esargs[0]= 0;
                      S->esatop= 0;
#ifdef MUDEBUG
                      S->esdlen= 0;
#endif
                      return ;
  }
}

static void print_chr(screen_t *S, uint32_t chr)
{
  int c;

  /** we need to handle the control characters here as well, because
      the UTF8 decoder can return them. **/
  if (S->escape_mode) { escape_chr(S, chr); return ; }
  if (chr<' ') { control_chr(S, chr); return ; }

  if (S->TXTFLAGS & TXTATTR_ALT)
  {
    if (chr>=' ' && chr<127) chr= altset[chr];
  }

  chr= ATTCHR(S->SGR, chr);
  if (S->pw)
  {
    if (S->m_wrap)
    {
      tIND(S);
      cursor(S, S->cursor_row, S->scroll_left);
    }
    else
    {
      S->pw= 0;
    }
  }

  if (S->m_insert) tICH(S, 1);
  c= S->cursor_col;
  MATRIX(S, S->cursor_row, c)= chr;
  c++;
  if (c>S->scroll_right) 
  {
    c= S->scroll_right;
    if (S->m_wrap) S->pw= 1;
  }
  // this also updates the old cursor position
  cursor(S, S->cursor_row, c); 
}


void screen_print_str(screen_t *S,const uint8_t *str, int len)
{
  int i,j, spc;
  const uint8_t *end;

  end= str + len;
restart:
  while(str<end && S->escape_mode) escape_chr(S, *str++);
  if (str==end) return ;
  if (str[0]<' ') { control_chr(S, *str++); goto restart; }
  if (!S->decoding && str[0]<127 && !S->m_insert &&
      (spc=S->ncols-S->cursor_col-1)>0)
  {
    for(i=0;i<spc && i+str<end;i++)
      if (str[i]<' ' || str[i]>=127)
         break;
    print_chrs_simple(S, str, i);
    str+= i;
  }
  while(str<end)
  {
    S->decoding= 1;
    if (utf8decoder_chr(&S->decoder, *str++))
    {
      for(j=0;j<S->decoder.outlen;j++) print_chr(S,S->decoder.outbuf[j]);
      utf8decoder_reset(&S->decoder);
      S->decoding= 0;
      goto restart;
    }
  }
}

void screen_set_fd(screen_t *S, int fd)
{
  S->out_fd= fd;
}

screen_t *screen_new(int nrows,int ncols)
{
  screen_t *S;
  int r;

  cell_space= ' ';
  cell_newline= '\n';

  init_altset();

  S= calloc(1,sizeof(*S));
  S->out_fd= -1;
  S->matrix_sel= 0;
  S->nrows= nrows; S->ncols= ncols;
  S->primary_matrix= malloc(ncols*nrows*sizeof(uint32_t));
  S->alternate_matrix= malloc(ncols*nrows*sizeof(uint32_t));
  fill_matrix(S, S->primary_matrix, ' ');
  fill_matrix(S, S->alternate_matrix, ' ');
  S->matrix= S->primary_matrix;
  S->saved_cursor= &S->primary_saved_cursor;

  S->tabstop= malloc(ncols * sizeof(S->tabstop[0]));
  tabstop_init(S);

  S->cursor_row= 0;
  S->cursor_col= 0;
  // set_cursor(0,0);
  S->m_wrap= 1;

  S->scroll_top= 0;  S->scroll_bot= S->nrows-1;
  S->scroll_left= 0; S->scroll_right= S->ncols-1;

  S->upd= malloc(sizeof(*S->upd)*nrows);
  for(r=0;r<nrows;r++)
    { S->upd[r].start= ncols; S->upd[r].end= -1; }
  S->updates= 0;
  S->rminupd= nrows-1;
  S->rmaxupd= 0;

  utf8decoder_init(&S->decoder, '?');

  update_all(S);

  S->last= calloc(1,sizeof(S->last[0]));
  S->last->prev= NULL;
  S->last->next= NULL;
  S->last->seqno= 1;
  S->last->ncols= S->ncols;
  S->last->size= S->last->len= S->ncols*S->nrows;
  S->last->data= S->primary_matrix;
  return S;
}


void screen_redraw(screen_t *S) 
{
  if (S->active)
  {
   if (S->m_chidden)
      set_cursor(-1,-1);
   else 
      set_cursor(S->cursor_row, S->cursor_col);  
  }
  update_all(S);
}






void screen_free(screen_t *S)
{
#if 0
  row_t *r;
  int i;
  if (S->mode)
  {
    for(i=0;i<S->nrows;i++) free_row(S->rows[i]);
    free(S->rows);
    free(S->STrows);
  }
  else
  {
    for(r=S->first;r;r=free_row(r))  ;
    free(S->rows);
  }
  free(S->temporows);
  free(S); 
#endif
}

void screen_activate(screen_t *S,int value)
{
  S->active= value;
}

void matrix_debug(int d)
{
  matrix_dbg= d;
}

int screen_mode(screen_t *S)
{
  return S->matrix_sel;
}

int screen_bs_mode(screen_t *S)
{
  return S->m_bs;
}

int screen_del_mode(screen_t *S)
{
  return S->m_del;
}

/**
  Scrollback Code
 **/

static void prt_row(screen_t *S, int r, pos_t *P)
{
  int c,i;
  int nc;
  uint32_t *D;
  uint32_t A;

  nc= imin2(P->buf->ncols, S->ncols);
  D= P->buf->data + P->index;
  c= P->index % P->buf->ncols;

  for(i=0;c<nc;c++,i++) display_glyph(r,c,D[i]);
  if (c<S->ncols)
  {
    A= i ? D[i-1] & ~CELL_CHRMASK : 0;
    for(;c<S->ncols;c++) display_glyph(r,c, A | ' ');
  }
}

static int fwd_row(pos_t *P)
{
  struct cbuf *B;
  int I;
  B= P->buf; I= P->index;
  I -= I % B->ncols;
  if (I+B->ncols<B->len)
  {
    P->index = I + B->ncols;
    return 0;
  }
  B= B->next;
  if (!B || !B) return 1;
  P->buf= B;
  P->index= 0;
  return 0;
}

static int bwd_row(pos_t *P)
{
  struct cbuf *B;
  int I;
  B= P->buf; I= P->index;
  if (I-B->ncols>=0)
  {
    P->index -= B->ncols;
    return 0;
  }
  B= B->prev;
  if (!B || B->len-B->ncols<0) return 1;
  P->buf= B;
  P->index= B->len-B->ncols;
  return 0;
}

static uint32_t pos2cindex(pos_t P)
{
  return P.buf->coff + P.index;
}

static uint32_t rc2cindex(screen_t *S, int r, int c)
{
  pos_t P;
  if (S->matrix_sel==1) return r*S->ncols+c;
  if (!S->vs) return r*S->ncols + c + S->last->coff;
  P= S->toppos;
  while(r--) if (fwd_row(&P)) break;
  c= umin2(c, P.buf->ncols);
  return P.buf->coff + P.index + c;
}

static void cindex2rc(screen_t *S,uint32_t IDX, int *Rr, int *Rc)
{
  // we are guaranteed to have IDX within the screen
  // we are also guaranteed to have a valid S->toppos because
  // this only gets called in scrollback mode.
  pos_t P;
  int row;
  P= S->toppos;
  row= 0;
  while(IDX-pos2cindex(P)>=P.buf->ncols)
    if (fwd_row(&P)) break;
    else row++;
  *Rr= row;
  *Rc= umin2(S->ncols, IDX-pos2cindex(P));
}

static void screen_updates_scrollback(screen_t *S)
{
  int urow;
  pos_t pos;
  int i;
  uint32_t B, E, topos, bopos; int sbr, sbc, ser, sec;

  if (!S->vs || S->vsrow<0 || S->vsrow>S->nrows) return ;
  if (S->sel_type!=SEL_MAIN) goto nosel;
  B= umin2(S->sel_begin, S->sel_end);
  E= umax2(S->sel_begin, S->sel_end);
  topos= pos2cindex(S->toppos);
  if (E<topos) goto nosel;
  bopos= rc2cindex(S,S->nrows-1,S->ncols-1);
  if (B>bopos) goto nosel;
  B= umax2(B, topos);
  E= umin2(E, bopos);
  cindex2rc(S, B, &sbr, &sbc);
  cindex2rc(S, E, &ser, &sec);
  set_sel(sbr, sbc, ser, sec);

draw:
  set_cursor(-1,-1);
  urow= S->vsrow;
  pos= S->updpos;
  for(i=0;i<S->vsnrow && urow<S->nrows;i++)
  {
    prt_row(S, urow++, &pos);
    if (fwd_row(&pos)) break;
  }
  update_area(0, S->vsrow, S->ncols, S->vsnrow);
  S->vsrow= -1;
  return ;
nosel:
  set_sel(-1,0,0,0);
  goto draw;
}

static void scrollback_top(screen_t *S, pos_t P)
{
  S->vs= 1;
  S->vsrow= 0;
  S->vsnrow= S->nrows;
  S->updpos= P;
  S->toppos= P;
  if (S->sel_type!=SEL_MAIN || !S->sel_state) return ;
  S->sel_end= rc2cindex(S, S->sel_row, S->sel_col);
  S->sel_state= 2;
}

void dbg_row(pos_t *P)
{
  uint32_t *D, *DE;
  D= P->buf->data + P->index;
  DE= D+P->buf->ncols;
  while(DE>D && (CELL_CHRMASK&DE[-1])==' ') DE--;
  while(D<DE) printf("%c", *D++ & 0xff);
  printf("\n");
}

void screen_page_prev(screen_t *S)
{
  pos_t P;
  int i;

  // this shouldn't be called with alternate buffer active
  if (S->matrix_sel) return ; 

  if (S->vs)
  {
    P= S->toppos;
  }
  else
  {
    P.buf= S->last;
    P.index= 0;
  }

  for(i=0;i<S->nrows-1;i++) if (bwd_row(&P)) break;
  if (!i) return ; // no changes, already at the top.

  if (P.buf==S->last) screen_page_last(S);
		 else scrollback_top(S, P);
}

void screen_page_next(screen_t *S)
{
  int i;
  pos_t P;
  // this shouldn't be called with alternate buffer active
  if (S->matrix_sel || !S->vs) return ; 
  P= S->toppos;
  for(i=0;i<S->nrows-1;i++) if (fwd_row(&P)) break;
  if (!i) return ;
  if (P.buf==S->last) screen_page_last(S);
                 else scrollback_top(S, P);
}

void screen_page_last(screen_t *S)
{
  if (!S->vs || S->matrix_sel) return ;
  S->vs= 0;
  if (S->active && !S->m_chidden) set_cursor(S->cursor_row, S->cursor_col);
  if (S->sel_type==SEL_MAIN && S->sel_state)
  {
    S->sel_end= rc2cindex(S, S->sel_row, S->sel_col);
    S->sel_state= 2;
  }
  update_all(S);
}



/*
  ┌──────────────────────────────────────────────────────────────────────────┐
  │                              SELECTION CODE                              │
  └──────────────────────────────────────────────────────────────────────────┘

  Here is the chain of events we're looking for.

  initially, B= E= invalid, state= 0

  from             to
  state   User    state   System action
  -----  ------   -----   ---------------
    0    Press      1     RESTART
    0    Move       0     -
    0    Release    0     -
    1    Press      1     RESTART
    1    Move       2     E=mpos 
    1    Release    0     B=inval
    2    Press      1     RESTART
    2    Move       2     E=mpos
    2    Release    0     Inform system about new selection. B= E= inval

 The RESTART action is defined as: 
   - clear the old selection
   - E= inval, B= mpos, state= 1
 This is done everytime a Press event is received. If it's received in
 a state that isn't consistent with our expectations, then we somehow
 missed some events. In this case I just restart the whole thing.

 When the mouse moves to another position, the new position is also included
 in the selection. If user wants to select a single character, then he needs
 to initially select two characters, and then walk it back to the single
 character. This is better than other approaches where you need to have
 either a mouse pointer shaped like a vertical bar and the user clicks 
 between columns or you lose the ability to select the whole screen at once.

 There are three modes of display, and each requires a different mechanism
 for handling selections.
  - alternate buffer
    This is the easiest one, all selections are on the alternate matrix
    which can be addressed only by <row,col> pairs.
  - main buffer, viewing the matrix
    A selection can span across the top of the screen. Other than that,
    this is pretty much the same as alternate buffer.
  - main buffer, viewing the scrollback 
    A selection is defined by a pair of buffer positions. When a change
    is done to a selection, the results should be displayed via the 
    scrollback update mechanism.

 There are other issues regarding selections.

 Ideally, all output from the application driving the terminal should be 
 suspended while there is an ongoing selection. This will make the user's
 life easier. However, if I do that, I also need to suspend keystrokes
 going to the application also. For this, I should buffer keystrokes 
 somehow and then replay them when the application is released after the
 selection is done. So, I won't do that. Instead, I will use an addressing
 mechanism which doesn't get so much affected by changes to the buffer.

 This will be the number of characters starting from the beginning of the
 scrollback buffer. I'll call these "cindex". 
 

 **/

static void update_continous(screen_t *S, int rs, int cs, int re, int ce)
{
  int r,c;
  for(r=rs,c=cs;r<=re;r++)
  {
    if (r==re) update_row(S, r, c, ce);
          else update_row(S, r, c, S->ncols-1);
    c= 0;
  }
}

static int udec(uint32_t *R, uint32_t V)
{
  if (*R<V) return 1;
  *R -= V;
  return 0;
}

 // update contents between given char indices..
static void update_cindex(screen_t *S, uint32_t B, uint32_t E)
{
  uint32_t it, ib;
  pos_t P;

  if (B>E) { uint32_t T; T= B; B= E; E= T; }
  if (S->matrix_sel==1)
  {
    update_continous(S, B/S->ncols, B%S->ncols, E/S->ncols, E%S->ncols);
    return ;
  }
  if (!S->vs)
  {
    if (udec(&E, S->last->coff)) return ;
    if (udec(&B, S->last->coff)) B= 0;
    update_continous(S, B/S->ncols, B%S->ncols, E/S->ncols, E%S->ncols);
    return ;
  }
  it= pos2cindex(S->toppos);
  if (E<it) return ;
  ib= rc2cindex(S, S->nrows-1, S->ncols-1);
  if (B>ib) return ;
  if (B<it) B= it;
  if (E>ib) E= ib;

  P= S->toppos;
  S->vsrow= 0;
  while(B-pos2cindex(P) >= P.buf->ncols)
    if (fwd_row(&P)) break;
    else {
     S->vsrow++;
     if (S->vsrow>=S->nrows)
     {
       // this should never happen because we already
       // clipped it above. but anyway, I will see if it
       // ever gets executed due to some offbyone mistake
       S->vsrow= -1;
       S->vsnrow= 0;
       return ;
     }
    }

  // now P and B point to the same row..
  P.index += umin2(B-pos2cindex(P),P.buf->ncols);
  // pos2cindex(P) SHOULD be equal to P now
  S->updpos= P;
  S->vsnrow= 1;
  while(E>pos2cindex(P))
    if (fwd_row(&P)) break;
    else S->vsnrow++;
}

static void screen_resetsel(screen_t *S)
{
  if (S->sel_type==S->matrix_sel+1)
     update_cindex(S, S->sel_begin, S->sel_end);
  S->sel_type= SEL_INVAL;
  S->sel_state= 0;
}

void screen_ppress(screen_t *S,int row,int col)
{
  screen_resetsel(S);
  S->sel_state= 1;
  S->sel_begin= rc2cindex(S, row, col);
  S->sel_col= col;
  S->sel_row= row;
}

void screen_pmove(screen_t *S,int row, int col)
{
  uint32_t oE;
  //
  // state is 1 or 2..
  //
  if (S->sel_state==0 || (S->sel_col==col && S->sel_row==row)) return ;
  if (S->sel_state==1) { oE= S->sel_begin; S->sel_type= S->matrix_sel+1; }
                  else oE= S->sel_end;
  S->sel_end= rc2cindex(S, row, col);
  update_cindex(S, oE, S->sel_end);
  S->sel_state= 2;
  S->sel_row= row;
  S->sel_col= col;
}

static pos_t cindex2pos(screen_t *S, uint32_t idx)
{
  pos_t P;
  struct cbuf *B;

  if (S->vs) B= S->toppos.buf;
        else B= S->last;
 
again:
  if (B->coff>idx)
  {
    if (B->prev) { B= B->prev; goto again; }
    P.index= 0;
  }
  else
  {
    if (idx-B->coff<B->len)
    {
      P.index= idx-B->coff;
    } 
    else
    {
      if (B->next) { B= B->next; goto again; }
      P.index= B->len-1;
    }
  }
  P.buf= B;
  return P; 
}

/** copies one row of data from the range [sp,ep] into dst!=NULL.
    *Roff is incremented by the amount. return value is 1
    if sp has not gone past ep yet. otherwise, it's 0. */
static int pos_copy (screen_t *S, pos_t *sp, pos_t *ep,
                     uint32_t *dst, uint32_t *Roff)
{
  struct cbuf *sb, *eb;
  int cp, scol, ecol;

  sb= sp->buf; eb= ep->buf;
  if (!sb || sb->coff>eb->coff) return 0;
  if (sb==eb) goto same_buffer;
advance_normal:
  scol= sp->index % sb->ncols;
  cp= 0;
  if (scol<S->ncols)
  {
    cp= imin2(S->ncols, sb->ncols)-scol;
    if (dst) 
       memcpy(dst + *Roff, sb->data+sp->index, cp*sizeof(*dst));
    *Roff += cp;
    scol += cp;
    if (dst)
      for(;scol<S->ncols;scol++) *(dst+*Roff++)= ' ';
    else
      *Roff += imax2(0, S->ncols - scol);
  }
  if (dst) *(dst+*Roff)='\n';
  *Roff += 1;
  sp->index += (sb->ncols-sp->index%sb->ncols);
  if (sp->index>=sb->len)
  {
    sp->buf= sb->next;
    sp->index= 0;
  }
  if (sp->buf) return 1;
  return 0;

same_buffer:
  if (sp->index>ep->index) return 0;
  if (sp->index/sb->ncols == ep->index/sb->ncols) goto same_row;
  goto advance_normal;

same_row:
  sp->buf= NULL; // do this for me..
  scol= sp->index % sb->ncols;
  if (scol>=S->ncols) return 1; // nothing to copy
  ecol= imin2(S->ncols-1, ep->index % eb->ncols);
  cp= ecol - scol + 1;
  if (dst) memcpy(dst + *Roff, sb->data+sp->index, cp * sizeof(*dst));
  *Roff+= cp;
  return 1;
}


void screen_prelease(screen_t *S)
{
  uint32_t B, E, sz;
  pos_t pB, pE, P;
  uint32_t *K, *O;
  int U,V,rb,re,cb,ce,r,c;

  if (S->sel_state!=2)
  {
    have_selection(NULL); // this will call select_none..
    return ;
  }
  if (S->sel_type!=S->matrix_sel+1)
  {
    // a press was done, followed by a release AFTER a mode change.
    // this doesn't make any sense.
    screen_resetsel(S);
    return ;
  }
  S->sel_state= 0;

  B= umin2(S->sel_begin, S->sel_end);
  E= umax2(S->sel_begin, S->sel_end);
  if (S->sel_type==SEL_ALT)
  {
    rb= B/S->ncols; re= E/S->ncols;
    cb= B%S->ncols; ce= E%S->ncols;
    O= K= malloc(sizeof(*K)*(E-B+1+re-rb+1));
    if (!K) { have_selection(NULL); return ; }
    for(r=rb,U=cb;r<=re;r++,U=0)
    {
      V= (r==re)? ce: S->ncols;
      for(c=U;c<V;c++)
        *O++= MATRIX(S, r, c) & CELL_CHRMASK;
      *O++= '\n';
    }
    *O= 0;
    have_selection(K);
    return ;
  }

  pB= cindex2pos(S, B);
  pE= cindex2pos(S, E);

  K= NULL;
all_over:
  P= pB;
  sz= 0;
  O= K;
  while(pos_copy(S,&P,&pE,K,&sz));
  if (!K)
  {
    K=malloc(sizeof(*K)*(sz+1));
    if (!K) { have_selection(NULL); return ; }
    goto all_over;
  }
  K[sz]= 0;
  have_selection(K);
}

void screen_select_none(screen_t *S)
{
  screen_resetsel(S);
}
