#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

#include "msg.h"



/** message format

  offset     size     field
       0        4     message size
       4        4     type
       8        4     ni
      12        4     ns
      16      ni*4    integer data
      *       ns*4    length data 
      *         *     string/blob data
**/


static int mencode_int(unsigned char *B, int o, uint32_t V)
{
  B[o++]= V >> 24;
  B[o++]= V >> 16;
  B[o++]= V >> 8;
  B[o++]= V;
  return o;
}


static void mchan_realloc_buf(struct mchan *C, int sz)
{
  if (C->bsiz >= sz) return ;
 
  C->bsiz = sz + sz/2;
  C->buf= realloc(C->buf, C->bsiz);
}

static void mencode(struct mchan *C, struct msg *M)
{
  int i, sz, o;
  sz= 0;
  for(i=0;i<M->ns;i++) sz += M->larg[i] + 4;
  sz += M->ni * 4;
  sz += 16;        // header
  mchan_realloc_buf(C, sz);
  C->len= sz;
  o = 0;
  o = mencode_int(C->buf , o, sz );
  o = mencode_int(C->buf , o, M->type);
  o = mencode_int(C->buf , o, M->ni);
  o = mencode_int(C->buf , o, M->ns);
  for(i=0;i<M->ni;i++) o= mencode_int(C->buf, o, M->iarg[i]);
  for(i=0;i<M->ns;i++) o= mencode_int(C->buf, o, M->larg[i]);
  for(i=0;i<M->ns;i++)
  {
    if (M->sarg[i]) memcpy(C->buf+o, M->sarg[i], M->larg[i]);
    o += M->larg[i];
  }
  C->pending_write= 1;
  C->woff= 0;
  if (M->encoded)
     (*M->encoded)(M);
}

static uint32_t mdecode_int(unsigned char *B)
{
  return ((uint32_t) B[0]<<24) 
       | ((uint32_t) B[1]<<16) 
       | ((uint32_t) B[2]<<8) 
       | ((uint32_t) B[3]);
}


static void mdecode(struct mchan *C, struct msg *M)
{
  int o, i;
  o= 4; // skip the size..
  M->type= mdecode_int(C->buf+o);   o+= 4;
  M->ni=   mdecode_int(C->buf+o);   o+= 4;
  M->ns=   mdecode_int(C->buf+o) ;  o+= 4;
  for(i=0;i<M->ni;i++,o+=4) M->iarg[i]= mdecode_int(C->buf+o);
  for(i=0;i<M->ns;i++,o+=4) M->larg[i]= mdecode_int(C->buf+o);
  for(i=0;i<M->ns;i++)
  {
    M->sarg[i]= (char*) C->buf + o;
    o += M->larg[i];
  }
  C->roff= 0;
  C->len= 0;
}



void  mchan_init   (struct mchan *C, int fd, int rW)
{
  memset(C, 0, sizeof(*C));
  C->fd= fd;
  C->rW= rW;
  C->bsiz= 1024;
  C->buf= malloc(C->bsiz);

  if (rW)
  {
    C->mbsiz= 16;
    C->mbuf= malloc(C->mbsiz*sizeof(C->mbuf[0]));
  }
}

void  mchan_free   (struct mchan *C)
{
  if (C->rW) free(C->mbuf);
  free(C->buf);
  C->fd= -1;
}

void  mchan_queue  (struct mchan *C, struct msg *M)
{
  struct msg *A;
  int ns, i;

  if (!C->pending_write)
  {
    mencode(C, M);
    return ;
  }
 
again:
  if (C->mblen<C->mbsiz)
  {
    C->mbuf[(C->mbhead+C->mblen++)%C->mbsiz]= *M;
    return ;
  }
 
  ns= C->mbsiz + C->mbsiz/2;
  A= malloc(sizeof(*A)*ns);
  for(i=0;i<C->mblen;i++)
    A[i]= C->mbuf[(C->mbhead+i)%C->mbsiz];
  C->mbhead= 0;
  free(C->mbuf);
  C->mbuf= A;
  goto again;
}

static int mchan_getb(struct mchan *C, int len)
{
  ssize_t R;
  if (C->fd<0) return 1;
  mchan_realloc_buf(C, len+C->roff);
again:
  R= recv(C->fd, C->buf+C->roff, len, MSG_DONTWAIT);
  if (R>=0)
  {
    C->roff += R;
    return 0;
  }
  switch(errno)
  {
#if EAGAIN != EWOULDBLOCK
    case EAGAIN: 
#endif
    case EWOULDBLOCK:
      return 0;
    case EINTR:
      goto again;
  }
  C->fd= -1;
  return 1;
}

static int mchan_putb(struct mchan *C)
{
  ssize_t R;
  if (C->fd<0) return 1;
again:
  R= send(C->fd, C->buf+C->woff, C->len-C->woff, MSG_DONTWAIT);
  if (R>=0)
  {
    C->woff += R;
    return 0;
  }
  switch(errno)
  {
#if EAGAIN != EWOULDBLOCK
    case EAGAIN: 
#endif
    case EWOULDBLOCK:
      return 0;
    case EINTR:
      goto again;
  }
  C->fd= -1;
  return 1;
}


int   mchan_recv   (struct mchan *C, struct msg *M)
{
  if (C->roff<4)
  {
    if (mchan_getb(C,4-C->roff)) return -1;
    if (C->roff<4) return 0;
    C->len= mdecode_int(C->buf);
    return 0; 
  }
  if (mchan_getb(C, C->len-C->roff)) return -1;
  if (C->roff<C->len) return 0;
  mdecode(C, M);
  C->roff= 0;
  return 1;
}

int   mchan_send   (struct mchan *C)
{
  if (C->pending_write) 
  {
    if (mchan_putb(C)) return -1;
    if (C->woff!=C->len) return 0;
    C->woff= 0;
    C->len= 0;
    C->pending_write= 0;
  }
  if (C->mblen==0) return 1;
  mencode(C, C->mbuf+C->mbhead);
  C->mbhead = (C->mbhead+1)%C->mbsiz;
  C->mblen--;
  return 0;
}

int socket_accept(int sfd)
{
  struct sockaddr_in caddr;
  socklen_t caddr_len;
  caddr_len= sizeof(caddr);
  return accept(sfd, (struct sockaddr*) &caddr, &caddr_len);
}

int socket_server(int port,int backlog)
{
  int sfd=-1,  one=1;
  int sn;
  static const char *step[]= { "socket", "setsockopt", "bind",
                               "listen" };
  struct sockaddr_in saddr;

  sn= 0;
  if ((sfd= socket(AF_INET, SOCK_STREAM, 0))<0) goto errout; 

  sn++;
  if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR | SO_REUSEPORT,
      &one, sizeof(one))) goto errout;

  sn++;
  saddr.sin_family= AF_INET;
  saddr.sin_addr.s_addr= INADDR_ANY;
  saddr.sin_port= htons(port);
  if (bind(sfd, (struct sockaddr*) &saddr, sizeof(saddr))) goto errout;

  sn++;
  if (listen(sfd, backlog<=0 ? 5 : backlog)) goto errout;

  return sfd;

errout:
  perror(step[sn]);
  if (sfd>=0) close(sfd);
  return -1;
}

struct connect_params
{
  int pipefd[2];
  int error_no;
  int fd;
  struct sockaddr_in saddr;
};

static int ptrio(int fd, int rW, struct connect_params **V)
{
  unsigned char *K;
  int o, sz;
  ssize_t R;
  K= (unsigned char*) V;
  sz= sizeof(void*);
  o= 0;
  while(o!=sz)
  {
    R= rW ? write(fd, K+o, sz-o) : read(fd, K+o, sz-o);
    if (R<=0)
    {
      if (errno==EINTR || errno==EAGAIN) continue;
      return 1;
    }
    o+= R;
  }
  return 0;
} 

static void *async_connect(void *arg)
{
  struct connect_params *P;

  P= arg;
  if (connect(P->fd, (struct sockaddr*) &P->saddr, sizeof(P->saddr)))
    P->error_no= errno;
  else
    P->error_no= 0;
  
  if (ptrio(P->pipefd[1], 1, &P))
  {
    // unable to write to the pipe. we need to close everything and exit.
    close(P->pipefd[1]);
    close(P->fd);
    free(P);
  }
  return NULL;
}



int socket_connect(int fd)
{
  struct connect_params *P;
  int Rfd, Rerrno;

  if (ptrio(fd, 0, &P))
  {
    // this is extremely unlikely, but now we have a memory & fd leak.
    // the peer thread was unable to send the whole pointer.
    // hopefully this will be detected by the sender and it will
    // free the necessary resources.
    close(fd); // this is pipefd[0]
    return -1;
  }

  Rfd= P->fd;
  Rerrno= P->error_no;
  close(P->pipefd[0]);
  close(P->pipefd[1]);
  free(P);
  if (Rerrno)
  {
    close(Rfd);
    errno= Rerrno;
    return -1;
  }
  return Rfd;
}

int socket_client(const char *text_address, int port)
{
  pthread_t tid;
  pthread_attr_t tatt;
  struct connect_params *P;
  int sn;
  static const char *step[]= {"pipe", "inet_pton", "socket", "pthread_create"};

  P= malloc(sizeof(*P));
  P->fd= -1;
  P->pipefd[0]= P->pipefd[1]= -1;
  P->error_no= 0;

  sn= 0;
  if (pipe(P->pipefd)) goto errout;

  sn++;
  P->saddr.sin_family= AF_INET;
  P->saddr.sin_port= htons(port);
  if (!inet_pton(AF_INET, text_address, &P->saddr.sin_addr)) goto errout;
 
  sn++;
  P->fd= socket(AF_INET, SOCK_STREAM, 0);
  if (P->fd<0) goto errout;

  sn++;
  pthread_attr_init(&tatt);
  pthread_attr_setdetachstate(&tatt, PTHREAD_CREATE_DETACHED);
  if (pthread_create(&tid, &tatt, async_connect, P)) goto errout;
  pthread_attr_destroy(&tatt);

  return P->pipefd[0];
errout:
  perror(step[sn]);
  switch(sn)
  {
    case 3: pthread_attr_destroy(&tatt); // pthread has failed
            close(P->fd);  
    case 2:                // socket has failed
    case 1:                // inet_pton has failed
            close(P->pipefd[0]);
            close(P->pipefd[1]);
    case 0:                // pipe has failed
            free(P);
         break;
  }
  return -1;
}


