/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdarg.h>
#include <sys/signalfd.h>
#include "gfx.h"
#include "ptyutil.h"

#include "msg.h"
#include "mconst.h"

enum { FD_CONN= 100, FD_DATA, FD_CHILD, TIMER_CONNECT, TIMER_HEARTBEAT };
static struct mchan mc_in, mc_out;
static int ro_conn, ro_sock= -1;
static int sifd;

static char *disp_addr= "127.0.0.1";

typedef struct
{
  int fd;
  pid_t child_pid;
  int present;
  char buf[1024];
} term_t;

static term_t tab[8];

static void msg_out(int type, int ni, int ns, void (*encf)(struct msg*), ...)
{
  struct msg M;
  va_list ap;
  int i;
  M.type= type;
  M.ni= ni;
  M.ns= ns;
  M.encoded= encf;
  if (ro_sock<0) return ;
  va_start(ap, encf);
  for(i=0;i<ni;i++)
    M.iarg[i]= va_arg(ap, int);
  for(i=0;i<ns;i++)
  {
    M.larg[i]= va_arg(ap, int);
    M.sarg[i]= va_arg(ap, char*);
    if (M.larg[i]<0) M.larg[i]= strlen(M.sarg[i])+1;
  }
  va_end(ap);
  gfx_stop_monitor_file(ro_sock, -1);
  gfx_monitor_file(ro_sock, FD_DATA, 1, 1, 1);
  mchan_queue(&mc_out, &M);
  gfx_timer_stop(TIMER_HEARTBEAT);
}

static void tryexec(const char *fi)
{
  fprintf(stderr,"trying %s\n",fi);
  execl(fi, fi, NULL);
}

static void activate(int tn)
{
  if (tab[tn].present) return ;

  tab[tn].fd= ptymaster();
  if (tab[tn].fd<0)
  {
    perror("ptymaster");
    return ;
  }
  tab[tn].child_pid= fork();
  if (tab[tn].child_pid==-1)
  {
    perror("fork");
    close(tab[tn].fd);
    return ;
  }
  if (tab[tn].child_pid)
  {
// printf("tab %d child= %d\n", tn, tab[tn].child_pid);
    gfx_monitor_file(tab[tn].fd, tn, 1, 0, 1);
    // screen_set_fd stuff
    tab[tn].present= 1;
    return ;
  }
  if (ptyslave(tab[tn].fd))
  {
    perror("ptyslave");
    return ;
  }
  setenv("TERM", "xterm", 1);
  tryexec("/bin/bash");
  tryexec("/usr/bin/bash");
  tryexec("/bin/sh");
  exit(1);
}

static void deactivate(int tn)
{
  if (!tab[tn].present) return ;
  if (tab[tn].fd<0) return ;
  
  msg_out(ROME_DEACTIVATE, 1, 0, NULL, tn);

  close(tab[tn].fd);
  gfx_stop_monitor_file(tab[tn].fd, -1);
  tab[tn].fd= -1;

  if (tab[tn].child_pid!=0)
  {
    waitpid(tab[tn].child_pid, NULL,0);
    tab[tn].child_pid= 0;
  }
  tab[tn].present= 0;
}

static void release_buf(struct msg *M)
{
  int tn;
  tn= M->iarg[0];
  if (tab[tn].fd>=0)
     gfx_monitor_file(tab[tn].fd, tn, 1, 0, 1);
}

static void read_tty(int tn)
{
  ssize_t R;

  if (tn<0 || tn>7) return ;

  if (tab[tn].fd<0)
  {
    deactivate(tn);
    return ;
  }

  if ((R= read(tab[tn].fd, tab[tn].buf, sizeof(tab[tn].buf)))<0
      && errno!=EINTR && errno!=EAGAIN) 
  {
    deactivate(tn);
  }
  if (R>0)
  {
    gfx_stop_monitor_file(-1, tn);
    msg_out(ROME_TTY2SCR, 1, 1, release_buf, tn, R, tab[tn].buf);
  }
}


static void mkconn()
{
  ro_conn= socket_client(disp_addr, 1605 );
  if (ro_conn<0) return ;
  gfx_monitor_file( ro_conn, FD_CONN, 1, 0, 0);
}

static int present_byte()
{
  int i, V;
  V= 0;
  for(i=0;i<8;i++) if (tab[i].present) V |= (1<<i);
  return V;
}

static void handle_connection()
{
  unsigned int w,h;
  ro_sock= socket_connect(ro_conn);
  gfx_stop_monitor_file( -1, FD_CONN);
  if (ro_sock<0) {
    gfx_timer_start( TIMER_CONNECT, 1500 );
    // printf("could not connect\n");
    return ;
  }
  mchan_init(&mc_in, ro_sock, 0);
  mchan_init(&mc_out, ro_sock, 1);
  msg_out(ROME_PRESENT, 1, 0, NULL, present_byte());
  gfx_geometry(&w, &h);
  msg_out(ROME_WINDOWDIM, 2, 0, NULL, w, h);
  gfx_monitor_file(ro_sock, FD_DATA, 1,1 ,1);
}


static void handle_except()
{
printf("socket closed by peer\n");
  close(ro_sock);
  mchan_free(&mc_in);
  mchan_free(&mc_out);
  gfx_stop_monitor_file(-1, FD_DATA);
  ro_sock= -1;
  mkconn();
}


static void handle_data()
{
  struct msg M;
  int R;
  int tn;

  switch(mchan_recv(&mc_in, &M))
  {
  case  0: return ;
  case  1: break;
  case -1: handle_except(); return ;
  }
  switch(M.type)
  {
  case ROME_ACTIVATE: activate(M.iarg[0]); break;
  case ROME_SCR2TTY:
    tn= M.iarg[0];
    if (tab[tn].fd<0) break;
    while(M.larg[0]>1)  // we don't write the NULL byte
    {
      R= write(tab[tn].fd, M.sarg[0], M.larg[0]-1);
      if (R<=0) break;
      M.sarg[0] += R;
      M.larg[0] -= R;
    }
    break;
   case ROME_REQCLIP:   gfx_request_x11_clipboard(); break;
   case ROME_REQSEL:    gfx_request_x11_selection(); break;
   case ROME_SEL2CLIP: gfx_copy_selection_to_clipboard(); break;
   case ROME_SETSEL:    gfx_set_x11_selection(strdup(M.sarg[0])); break;
   case ROME_RESETSEL:  gfx_reset_selection(); break;
  }
}
static void handle_out()
{
  if (ro_sock<0)
  {
    gfx_stop_monitor_file(-1, FD_DATA);
    return ;
  }
  switch(mchan_send(&mc_out))
  {
  case 0:
    break;
  case 1: 
    gfx_stop_monitor_file( -1, FD_DATA );
    gfx_monitor_file( ro_sock, FD_DATA, 1, 0, 1);
    gfx_timer_start(TIMER_HEARTBEAT, 5000);
    break;
  case -1:
    handle_except();
    break;
  }
}


void handle_dead_child()
{
  struct signalfd_siginfo si;
  ssize_t R;
  int i;
  R= read(sifd, &si, sizeof(si));
  if (R!=sizeof(si) || si.ssi_signo!=SIGCHLD) return ;
  for(i=0;i<8;i++)
    if (tab[i].child_pid==si.ssi_pid)
    {
      deactivate(i);
      break;
    }
}

void event_loop()
{
  gfx_event_t ev;
  int quit= 0;
  while(!quit)
  {
    if (!gfx_event(&ev)) continue;
    switch(ev.type)
    {
    case GFXEV_QUIT: quit= 1; break;
    case GFXEV_KEY_PRESS:
         msg_out(ROME_KEY, 2, 1, NULL, ev.key, ev.wide, -1,
                 ev.key==GFX_K_PRT ? ev.string : "");
         break;
    case GFXEV_READY_READ:
         switch(ev.key)
         {
         case FD_CONN: handle_connection(); break;
         case FD_DATA: handle_data(); break;
         case FD_CHILD: handle_dead_child(); break;
         case 0: case 1: case 2: case 3:
         case 4: case 5: case 6: case 7:
            read_tty(ev.key);
            break;
         }
         break;
    case GFXEV_READY_WRITE: if (ev.key==FD_DATA) handle_out(); break;
    case GFXEV_READY_EXCEPT:
         if (ev.key==FD_DATA) handle_except();
         break;
    case GFXEV(TIMER_CONNECT,TIMEOUT):
         mkconn();
         break;
    case GFXEV(TIMER_HEARTBEAT, TIMEOUT):
         msg_out(ROME_BEAT, 0, 0, NULL);
         gfx_timer_start(TIMER_HEARTBEAT, 5000);
         break;
    case GFXEV(Wplane, POINTER_MOVE):
         msg_out(ROME_POINTER, 2, 0, NULL, ev.x, ev.y);
         break;
    case GFXEV(Wplane, BUTTON_PRESS):                                           
         msg_out(ROME_CLICK, 3, 0, NULL, ev.button, ev.x, ev.y);
         break;
    case GFXEV(Wplane, BUTTON_RELEASE):
         msg_out(ROME_RELEASE, 3, 0, NULL, ev.button, ev.x, ev.y);
         break; 
    case GFXEV_SELECTION_CLEAR: msg_out(ROME_SELCLR, 0,0, NULL); break;
    case GFXEV_SELECTION_DATA: 
         msg_out(ROME_SELDATA, 0, 1, NULL, -1, ev.string);
         break;
    }
  }
}


int main(int argc,char **argv)
{
  sigset_t set;
  int opt;
  sigemptyset(&set);
  sigaddset(&set, SIGCHLD);
  sigprocmask(SIG_BLOCK, &set, NULL);
  sifd= signalfd(-1, &set, SFD_CLOEXEC);
  
  signal(SIGPIPE, SIG_IGN);

  while((opt=getopt(argc,argv,"h:"))!=-1) switch(opt)
  {
  case 'h': disp_addr= optarg; break;
  default: fprintf(stderr,"usage: %s [-h host]\n", argv[0]);
           return 1;
  }

  if (gfx_preinit(NULL)) return 1;

  mkconn();

  // gfx_set_window_size(fontW*wd_cols, fontH*(wd_rows+1));
  gfx_init(NULL);
  gx.screen(Smain);
  gfx_monitor_file(sifd, FD_CHILD, 1, 0, 0);
  gfx_pixgrid_begin_report_mouse_position(Wplane);

  event_loop();

  return 0;
}
